# This code is part of KQCircuits
# Copyright (C) 2025 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).

from kqcircuits.pya_resolver import pya
from kqcircuits.util.instance_hierarchy_helpers import formatted_cell_instance_hierarchy, get_cell_instance_hierarchy


def all_cell_parents(layout: pya.Layout, cell: pya.Cell) -> list[int]:
    """Finds all parent cell indices of a given cell, including cell itself

    Args:
        layout: The layout
        cell: The cell to find parents of

    Returns: list of cell indices of parent cells, where the first element is always ``cell`` itself.
    """
    to_do = list(cell.each_parent_cell())
    result = [cell.cell_index()]
    while len(to_do) > 0:
        c = to_do.pop(0)
        result.append(c)
        to_do.extend(layout.cell(c).each_parent_cell())
    return result


def find_cells_with_error(
    layout: pya.Layout, parent_cell_index: int | None = None
) -> list[tuple[int, str, pya.DPoint]]:
    """Find cell indices and error messages for all cells in the layout that have called the ``raise_error_on_cell``
    method.

    Args:
        layout (pya.Layout): The layout.
        parent_cell_index: if not None, return only errors where the given cell index is a parent

    Returns: List of (cell_index, error_message, error_position) tuples.
    """
    # Cells with errors have a TEXT cell instance inside them, which has the error message set as instance property.
    if parent_cell_index is not None and not layout.is_valid_cell_index(parent_cell_index):
        raise ValueError("parent_cell_index does not exist in the layout")

    result = []
    for cell in layout.each_cell():
        for inst in cell.each_inst():
            error_message = inst.property("error_on_cell")
            if error_message is not None:
                if parent_cell_index is None or parent_cell_index in all_cell_parents(layout, cell):
                    error_position = inst.property("error_on_cell_position")
                    if error_position is not None:
                        x, y = map(float, error_position.split(","))
                    else:
                        x, y = 0.0, 0.0
                    position = pya.DPoint(x, y)
                    result.append((cell.cell_index(), error_message, position))
                    continue  # No need to search any remaining instances
    return result


def formatted_errors_on_cells(layout: pya.Layout, parent_cell_index: int | None = None) -> str:
    """Generate a formatted string describing all errors raised on cells and where in the cell hierarchy these appear.

    Args:
        layout: The layout
        parent_cell_index: if not None, include only errors where the given cell index is a parent

    Returns: Formatted multiline string
    """
    result = ""
    for cell_index, error_message, error_position in find_cells_with_error(layout, parent_cell_index):
        instance_data = get_cell_instance_hierarchy(layout, cell_index)
        result += f"{error_message}\n"
        for inst_data in instance_data:
            result += f"Error instance at {str(inst_data.trans * error_position)}:\n"
            result += formatted_cell_instance_hierarchy(inst_data) + "\n"
    return result



# This code is part of KQCircuits
# Copyright (C) 2025 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).
import ast
import logging
from math import ceil
from typing import Any

from klayout import pya
from kqcircuits.simulations.cross_section_simulation import CrossSectionSimulation
from kqcircuits.simulations.simulation import Simulation
from kqcircuits.util.parameters import Param, pdt, add_parameters_from


@add_parameters_from(Simulation, "tls_layer_thickness", "tls_layer_material")
class CutSimulation(CrossSectionSimulation):
    """Class for cross-section simulations with geometry generated from a 3D Simulation.

    By default, renames the layers such that partition regions of source simulation are basically ignored.
    The ``region_map`` can be used to define new partition regions or copy partition regions from source simulation.

    Adds interface layers 'ma', 'ms', and 'sa' if ``tls_layer_thickness`` and ``tls_layer_material`` are set properly.
    """

    source_sim = Param(pdt.TypeNone, "Instance of Simulation subclass from which the cross section is generated", None)
    cut_start = Param(pdt.TypeShape, "Start point of the cut segment", pya.DPoint(0, 0))
    cut_end = Param(pdt.TypeShape, "End point of the cut segment", pya.DPoint(100, 0))
    cut_bottom = Param(pdt.TypeDouble, "Bottom of the cross section", -1e30, unit="µm")
    cut_top = Param(pdt.TypeDouble, "Top of the cross section", 1e30, unit="µm")
    region_map = Param(
        pdt.TypeString,
        "Dictionary with new partition region suffixes as keys and region definitions as values",
        "{}",
        docstring="The region definition is a list of pya.DBox or string instances. Use string instance to copy "
        "partition region from the source simulation.",
    )
    metal_height = Param(pdt.TypeDouble, "Thickness of metal layer if sheet in the source simulation.", 0.2, unit="µm")

    def build(self):
        self.insert_cross_section_layers()
        partition_regions = self.get_partition_regions()
        self.process_geometry()
        self.apply_partition_regions(partition_regions)

    def process_geometry(self):
        self.regroup_layers()
        self.insert_tls_layers()

    def limited_region(self, box: pya.DBox):
        bottom = max(box.bottom, self.cut_bottom)
        top = min(box.top, self.cut_top)
        if bottom >= top:
            return pya.Region()
        return pya.Region(pya.DBox(box.left, bottom, box.right, top).to_itype(self.layout.dbu))

    def insert_cross_section_layers(self):
        """Insert cross sections from each layer of ``self.source_sim`` into ``self.layers``.
        Thicken sheet metals by ``self.metal_height``.
        """
        layout = self.source_sim.layout
        cut_edge = pya.DEdge(self.cut_start, self.cut_end).to_itype(layout.dbu)
        cut_vector = cut_edge.d()

        # Place constants related to non-orthogonal edges warning
        appr_edge_slope_tolerance = 0.2  # warning is given if edge slope compared to orthogonal exceeds appr. this
        database_unit_tolerance = 2  # the database unit tolerance

        # Compute variables related to non-orthogonal edges warning
        cut_region_width = ceil(0.5 * database_unit_tolerance / appr_edge_slope_tolerance) * 2
        max_cut_vector_sprod = cut_region_width * appr_edge_slope_tolerance * cut_vector.abs()

        # Simple path region for cut with small width. Use KLayout's boolean operators to detect the intersections.
        cut_region = pya.Region(pya.Path([cut_edge.p1, cut_edge.p2], cut_region_width))

        # Scale intersection dot products within confines of the cut
        crossing_edges = [e for s in cut_region.each() for e in s.each_edge() if cut_edge.crossed_by(e)]
        prods = [cut_vector.sprod(cut_edge.crossing_point(e)) for e in crossing_edges]
        cut_min = min(prods)
        cut_length = (self.cut_end - self.cut_start).abs()
        cut_scale = cut_length / (max(prods) - cut_min)

        regions = {}
        sheet_metals = {}
        for name, data in self.source_sim.layers.items():
            if "layer" not in data:
                segments = [(0.0, cut_length)]
            else:
                layer_region = pya.Region(self.source_sim.cell.begin_shapes_rec(layout.layer(data["layer"], 0)))
                intersection = (cut_region & layer_region).merged()
                segments = []
                for polygon in intersection.each():
                    crossing_edges = [e for e in polygon.each_edge() if cut_edge.crossed_by(e)]

                    # Warn if cross-section is taken with non-orthogonal edges
                    skew_edges = [e for e in crossing_edges if abs(cut_vector.sprod(e.d())) > max_cut_vector_sprod]
                    for skew_edge in skew_edges:
                        logging.warning(
                            f"Cross section is taken with non-orthogonal edge from simulation '{self.source_sim.name}' "
                            f"layer '{name}' at location ({cut_edge.crossing_point(skew_edge).to_dtype(layout.dbu)})."
                        )

                    # Calculate intersection as value between 0 and cut length
                    dists = [
                        (cut_vector.sprod(cut_edge.crossing_point(e)) - cut_min) * cut_scale for e in crossing_edges
                    ]
                    segments.append((min(dists), max(dists)))

            if not segments:
                continue

            # Add sheet metal cross sections to sheet_metals dictionary (to insert them later)
            if data["thickness"] == 0 and "excitation" in data:
                edges = [pya.DEdge(left, data["z"], right, data["z"]) for left, right in segments]
                sheet_metals[name] = {
                    "edges": pya.Edges([e.to_itype(self.layout.dbu) for e in edges]),
                    "material": data["material"],
                    "excitation": data["excitation"],
                }
                continue

            # Build region out of segments
            bottom, top = data["z"], data["z"] + data["thickness"]
            regions[name] = pya.Region()
            for left, right in segments:
                regions[name] += self.limited_region(pya.DBox(left, bottom, right, top))

            # Apply subtractions defined by layers
            for subtract in data.get("subtract", []):
                if subtract in regions:
                    regions[name] -= regions[subtract]

            # Insert layer that has material and non-empty region
            if data.get("material") is not None:
                excitation = {"excitation": data["excitation"]} if "excitation" in data else {}
                self.insert_layer(name, regions[name], data["material"], **excitation)

        self.insert_sheet_metals(sheet_metals)

    def insert_sheet_metals(self, sheet_metals: dict[str, dict[str, Any]]):
        """Inserts sheet metal layers thickened by ``self.metal_height`` on vacuum side."""
        if self.metal_height <= 0.0:
            return

        vacuum_regions = [d["region"] for d in self.layers.values() if d.get("material") == "vacuum"]
        for name, data in sheet_metals.items():
            extended_region = data["edges"].extents(0, round(self.metal_height / self.layout.dbu))
            region = pya.Region()
            for vacuum_region in vacuum_regions:
                region += extended_region & vacuum_region
                vacuum_region -= extended_region

            data = self.source_sim.layers[name]
            self.insert_layer(name, region, **{k: v for k, v in data.items() if k != "edges"})

    def get_partition_regions(self) -> dict[str, pya.Region]:
        """Return partition regions defined by ``self.region_map``.

        Returns:
             dictionary containing the layer name suffixes as keys and regions as values.
        """
        region_map = ast.literal_eval(self.region_map) if isinstance(self.region_map, str) else self.region_map
        regions = {}
        for part, definitions in region_map.items():
            regions[part] = pya.Region()

            for definition in definitions:
                if isinstance(definition, pya.DBox):
                    regions[part] += self.limited_region(definition)
                elif isinstance(definition, str):
                    for name, data in self.layers.items():
                        if name.endswith(definition):
                            regions[part] += data["region"].extents()  # add bboxes of each polygon of data["region"]
                            # The above could be relaxed to regions[part] += data["region"] if ma_layer was extended
                            # from metal to vacuum and not the opposite.
                else:
                    logging.warning(f"The region_map term {definition} is ignored due to unsupported type.")

            regions[part].merge()
            if regions[part].is_empty():
                logging.warning(f'Partition region "{part}" in 2D CutSimulation "{self.name}" is empty.')

        return regions

    def apply_partition_regions(self, regions: dict[str, pya.Region]):
        """Partition every non-metal layer into regions.

        Args:
            regions: dictionary containing the layer name suffixes as keys and regions as values.
        """
        for name, data in list(self.layers.items()):
            if "excitation" in data:
                continue
            for part, region in regions.items():
                self.insert_layer(f"{name}{part}", **{**data, "region": region & data["region"]})
                data["region"] -= region

    def regroup_layers(self):
        """Group and rename layers such that a layer name doesn't start other layer name.

        For example, if there are two source layers 'vacuum', 'vacuum_part', then both are merged into 'vacuum' layer.
        """
        old_layers = self.layers.copy()
        self.layers.clear()
        base_names = [n for n in old_layers if not any(n.startswith(k) for k in old_layers if n != k)]

        for name, data in old_layers.items():
            base_name = next(n for n in base_names if name.startswith(n))
            if base_name in self.layers:
                self.layers[base_name]["region"] += data["region"]
                base_wo_region = {k: v for k, v in self.layers[base_name].items() if k != "region"}
                data_wo_region = {k: v for k, v in data.items() if k != "region"}
                if base_wo_region != data_wo_region:
                    raise ValueError(f"Inconsistent layer data for '{base_name}'.")
            else:
                self.insert_layer(base_name, **data)

    def insert_tls_layers(self):
        """Insert TLS interface layers into the model."""
        metals, vacuums, substrates = self.get_metals_vacuums_substrates()
        names = ["ma_layer", "ms_layer", "sa_layer"]
        sources = [vacuums, metals, vacuums]
        targets = [metals, substrates, substrates]
        interfaces = [self.sum_region(s).edges() & self.sum_region(t).edges() for s, t in zip(sources, targets)]
        thickness = [float(Simulation.ith_value(self.tls_layer_thickness, i)) for i in range(3)]
        material = [Simulation.ith_value(self.tls_layer_material, i) for i in range(3)]

        for i in range(3):
            extension = round(thickness[i] / self.layout.dbu)
            region = interfaces[i].extended(0, 0, extension, 0, True) & self.sum_region(targets[i])
            self.insert_layer(names[i], region, material[i])
            self.subtract_region(targets[i], region)

    def get_metals_vacuums_substrates(self) -> tuple[list[str], list[str], list[str]]:
        """Return layer names for metals, vacuums, and substrates."""
        metals = [n for n, d in self.layers.items() if "excitation" in d]
        vacuums = [n for n, d in self.layers.items() if d.get("material") == "vacuum"]
        substrates = [n for n in self.layers if n not in metals + vacuums]
        return metals, vacuums, substrates

    def sum_region(self, layers):
        """Return union of regions of given layers."""
        region = pya.Region()
        for layer in layers:
            region += self.layers[layer]["region"]
        return region

    def subtract_region(self, layers, region):
        """Subtract region from given layers."""
        for layer in layers:
            self.layers[layer]["region"] -= region

    # pylint: disable=no-value-for-parameter
    def get_material_dict(self):
        """Override CrossSectionSimulation method to combine source simulation and self materials"""
        return {**self.source_sim.get_material_dict(), **super().get_material_dict()}

    def get_parameters(self):
        """Override CrossSectionSimulation method to combine source simulation and self parameters"""
        return {
            **super().get_parameters(),
            "source_sim": self.source_sim.__class__.__name__,
            **{f"source_{k}": v for k, v in self.source_sim.get_parameters().items()},
        }

    # pylint: enable=no-value-for-parameter



# This code is part of KQCircuits
# Copyright (C) 2025 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).


def refine_metal_edges(
    size: float = 5.0,
    slope: float = 0.3,
    excitation: int = -1,
    region: str = "",
    ignore_region: str = "",
    face_id: str = "*",
):
    """Returns the mesh_size dictionary to refine the Gmsh mesh at base metal edges.

    This function is useful for most Elmer 3d simulations. The field usually concentrates near the base metal edges, so
    it's beneficial to refine the mesh there.

    The default value for slope argument is chosen to be suitable for typical capacitance or EPR simulations.
    Make sure that the size value match the requirements of your simulation.

    Typical usage is to pass parameter mesh_size=refine_metal_edges(size) for a 3d ElmerSolution. Do not use this
    function with ElmerCrossSectionSolution.

    Args:
        size: mesh element length at base-metal edges [µm]
        slope: determines how fast the mesh element length can increase outside the edge [µm/µm] (0 < slope <= 1)
        excitation: metal excitation value where the refinement is applied, or -1 to cover all excitations
        region: partition region name in which the refinement is applied, or empty string to cover all regions
        ignore_region: partition region name in which the refinement is ignored, or empty string to not use limitation
        face_id: the string face id where the refinement is applied, or * to cover all faces

    """

    metals = ["ground", "signal_*"] if excitation < 0 else ["ground" if excitation == 0 else f"signal_{excitation}"]

    intersections = [f"{face_id}_gap"]
    if region:
        intersections.append(f"*{region}")
    if ignore_region:
        intersections.append(f"!*{ignore_region}")

    return {"&".join([f"{face_id}_{metal}"] + intersections): [size, size, slope] for metal in metals}



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).


import abc

import logging

from kqcircuits.elements.element import Element
from kqcircuits.pya_resolver import pya
from kqcircuits.simulations.simulation import Simulation, get_simulation_layer_by_name
from kqcircuits.util.geometry_helper import merge_points_and_match_on_edges
from kqcircuits.util.parameters import add_parameters_from


@add_parameters_from(Simulation, "name", "material_dict", "extra_json_data")
class CrossSectionSimulation:
    """Base class for cross-section simulation geometries.

    The geometry consist of 2-dimensional layers which are thought as cross sections of 3-dimensional geometry.
    This allows for modeling waveguide-like geometries where the cross section is constant for relatively long segments.

    This class is intended to be subclassed by a specific simulation implementation; The subclass should implement the
    method 'build' which defines the simulation geometry and material properties. The helper function 'insert_layer'
    should be called internally to build the geometry.
    """

    LIBRARY_NAME = None  # This is needed by some methods inherited from Element.

    def __init__(self, layout, **kwargs):
        """Initialize a CrossSectionSimulation.

        The initializer parses parameters, creates a top cell, and then calls `self.build` to create
        the simulation geometry.

        Args:
            layout: the layout on which to create the simulation

        Keyword arguments:
            `**kwargs`:
                Any parameter can be passed as a keyword argument.
        """
        if layout is None or not isinstance(layout, pya.Layout):
            error_text = "Cannot create simulation with invalid or nil layout."
            error = ValueError(error_text)
            logging.error(error_text)
            raise error

        self.layout = layout
        schema = type(self).get_schema()

        # Apply kwargs or default value
        for parameter, item in schema.items():
            if parameter in kwargs:
                setattr(self, parameter, kwargs[parameter])
            else:
                setattr(self, parameter, item.default)

        self.cell = kwargs["cell"] if "cell" in kwargs else layout.create_cell(self.name)

        self.layers = {}
        self.units = kwargs.get("units", "um")
        self.build()
        self.process_layers()

    # Inherit specific methods from Element
    get_schema = classmethod(Element.get_schema.__func__)
    get_layers = Simulation.get_layers
    get_parameters = Simulation.get_parameters
    is_metal = Simulation.is_metal
    get_material_dict = Simulation.get_material_dict
    check_material_dict = Simulation.check_material_dict

    @abc.abstractmethod
    def build(self):
        """Build simulation geometry.

        This method is to be overridden, and the overriding method should create the geometry to be simulated.
        """
        return

    def process_layers(self):
        """Process and check validity of self.layers. Is called after build method.

        Internally,
        * call merge_points_and_match_on_edges to avoid small-scale geometry artifacts
        * ignore layers with empty region
        * check if 'excitation' keywords are set correctly
        * insert layer shapes to self.cell.shapes and replace 'region' keyword with 'layer' keyword.
        """
        merge_points_and_match_on_edges([layer["region"] for layer in self.layers.values()], tolerance=1)

        layers = {}
        for name, data in self.layers.items():
            if data["region"].is_empty():
                continue  # ignore layer with empty region

            # check if excitation keyword is set correctly
            if self.is_metal(data.get("material")):
                if "excitation" not in data:
                    raise ValueError(f"Layer {name} is metal but integer excitation value is not set.")
                if not isinstance(data["excitation"], int):
                    raise ValueError(f"Excitation should be integer but {data['excitation']} is set for layer {name}.")
            elif "excitation" in data:
                raise ValueError(f"Layer {name} is dielectric but excitation value is set.")

            # insert layer shape to self.cell.shapes and replace region keyword with layer keyword
            layer_info = get_simulation_layer_by_name(name)
            self.cell.shapes(self.layout.layer(layer_info)).insert(data["region"])
            layers[name] = {"layer": layer_info.layer, **{k: v for k, v in data.items() if k != "region"}}

        self.layers = layers

    def insert_layer(self, layer_name: str, region: pya.Region, material: str, **params):
        """Add layer parameters into 'self.layers' if region is non-empty."""
        if not region.is_empty():
            self.layers[layer_name] = {"region": region.dup(), "material": material, **params}

    def restrict_layer_regions(self, bbox: pya.DBox):
        """Limit the regions of self.layers inside the bounding box."""
        region = pya.Region(bbox.to_itype(self.layout.dbu))
        for layer in self.layers.values():
            layer["region"] &= region

    def get_unfilled_region(self, bbox: pya.DBox) -> pya.Region:
        """Return region inside bbox that is not covered yet by layers."""
        region = pya.Region(bbox.to_itype(self.layout.dbu))
        for layer in self.layers.values():
            region -= layer["region"]
        return region

    def get_simulation_data(self):
        """Return the simulation data in dictionary form.

        Returns:
            dictionary of relevant parameters for simulation
        """
        simulation_data = {
            "simulation_name": self.name,
            "units": self.units,
            "layers": self.layers,
            "material_dict": self.check_material_dict(),
        }
        return simulation_data



# This code is part of KQCircuits
# Copyright (C) 2025 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).


from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.junctions.junction import Junction
from kqcircuits.util.symmetric_polygons import polygon_with_vsym


class SuperInductor(Junction):
    """The PCell declaration for a Manhattan style single junction."""

    junction_length = Param(pdt.TypeDouble, "Junction length", 2, unit="μm")
    wire_width = Param(pdt.TypeDouble, "Wire width", 0.02, unit="μm")
    phase_slip_junction_length = Param(pdt.TypeDouble, "Phase slip junction length", 0.5, unit="μm")
    squid_area_width = Param(pdt.TypeDouble, "Width of Squids Area.", 2, unit="μm")
    squid_area_height = Param(pdt.TypeDouble, "Height of Squids Area.", 5, unit="μm")
    squid_count = Param(pdt.TypeInt, "Number of Squids in the Super Inductor.", 8)
    squid_wire_length = Param(pdt.TypeDouble, "Length of the wire between SQUIDs.", 0.25, unit="μm")
    squid_x_connector_offset = Param(
        pdt.TypeDouble,
        "Length of the horizontal connector between the corner of the junction to the wire of the squid.",
        0.5,
        unit="μm",
    )
    inductor_padding = Param(pdt.TypeDouble, "Vertical margin from bends to limit of the circuit", 1, unit="μm")
    taper_horizontal_displacement = Param(
        pdt.TypeDouble,
        "Horizontal displacement of the taper from the junction.",
        1,
        unit="μm",
    )
    taper_horizontal_displacement_step = Param(
        pdt.TypeDouble,
        "Horizontal displacement step of the taper from the junction.",
        0.25,
        unit="μm",
    )
    bend_width_multiplier = Param(
        pdt.TypeDouble,
        "Relative width of bend according to junction_width",
        1.5,
    )
    shadow_width = Param(pdt.TypeDouble, "Width of the shadow layer.", 0.125, unit="μm")
    shadow_min_height = Param(pdt.TypeDouble, "Minimum height of the shadow layer.", 0.04, unit="μm")
    include_base_metal_gap = Param(pdt.TypeBoolean, "Include base metal gap layer.", True)
    include_base_metal_addition = Param(pdt.TypeBoolean, "Include base metal addition layer.", True)
    shadow_margin = Param(pdt.TypeDouble, "Shadow layer margin near the the pads.", 0.5, unit="μm")
    finger_overlap = Param(pdt.TypeDouble, "Length of fingers inside the pads.", 1.0, unit="μm")
    inductor_height = Param(pdt.TypeDouble, "Height of the junction element.", 22.0, unit="μm")
    inductor_width = Param(pdt.TypeDouble, "Width of the junction element.", 22.0, unit="μm")

    def build(self):
        self.produce_super_inductor()

    def produce_super_inductor(self):
        # create rectangular junction-support structures and junctions
        self._make_super_inductor_junctions()
        self._produce_ground_metal_shapes()
        self._produce_ground_grid_avoidance()
        self._add_refpoints()

    def _make_super_inductor_junctions(self):
        layer_name = "SIS_junction"
        junction_shapes = self.cell.shapes(self.get_layer(layer_name))
        shadow_shapes = self.cell.shapes(self.get_layer("SIS_shadow"))

        def wire_shadow(height, i, x, y):
            shadow_shape = pya.DBox(0, 0, self.shadow_width, height)
            side = (self.wire_width) if i % 2 == 0 else -self.shadow_width
            shadow_transform = pya.DTrans(0, False, (x + side), y)
            return shadow_shape * shadow_transform

        def junction_shadow(width, height):
            shadow_shape = pya.DBox(0, -self.shadow_min_height / 2, self.shadow_width, self.shadow_min_height / 2)
            left_transform = pya.DTrans(0, False, -self.shadow_width, height / 2)
            right_transform = pya.DTrans(0, False, width, height / 2)
            return [
                left_transform * shadow_shape,
                right_transform * shadow_shape,
            ]

        # SQUID ARRAY - Corrected version
        squid_width = self.squid_area_width
        squid_count = self.squid_count
        gap_size = self.squid_wire_length

        # Calculate available space for the boxes themselves
        total_gap_space = gap_size * (squid_count - 1)
        available_space_for_boxes = self.squid_area_height - total_gap_space
        squid_height = available_space_for_boxes / squid_count
        squid_offset_x = (-squid_width / 2) + self.squid_x_connector_offset - (self.wire_width / 2)
        squid_offset_y = self.inductor_height / 2 - self.squid_area_height / 2
        squid_transform = pya.DTrans(0, False, squid_offset_x, squid_offset_y)

        # Squid Junctions
        for i in range(int(squid_count)):
            shape_lower = pya.DBox(0, 0, squid_width, self.junction_width)
            y_position = i * (squid_height + gap_size)
            squid_lower_transform = pya.DTrans(0, False, 0, y_position)
            squid_upper_transform = pya.DTrans(0, False, 0, (squid_height - self.junction_width) + y_position)
            junction_shapes.insert(squid_lower_transform * shape_lower * squid_transform)
            junction_shapes.insert(squid_upper_transform * shape_lower * squid_transform)
            [left, right] = junction_shadow(squid_width, self.junction_width)
            shadow_shapes.insert(squid_lower_transform * left * squid_transform)
            shadow_shapes.insert(squid_lower_transform * right * squid_transform)
            shadow_shapes.insert(squid_upper_transform * left * squid_transform)
            shadow_shapes.insert(squid_upper_transform * right * squid_transform)

        # Squid Pylon Wires
        for i in range(int(squid_count)):
            x_offset = self.squid_x_connector_offset - (self.wire_width / 2)
            shape_left = pya.DBox(x_offset, 0, x_offset + self.wire_width, squid_height)
            shape_right = pya.DBox(squid_width - x_offset - self.wire_width, 0, squid_width - x_offset, squid_height)
            y_position = i * (squid_height + gap_size)
            junction_shapes.insert(pya.DTrans(0, False, 0, y_position) * shape_left * squid_transform)
            junction_shapes.insert(pya.DTrans(0, False, 0, y_position) * shape_right * squid_transform)
            ws_left = wire_shadow(
                height=squid_height - (self.junction_width * 2), i=1, x=x_offset, y=y_position + self.junction_width
            )
            shadow_shapes.insert(ws_left * squid_transform)
            ws_right = wire_shadow(
                height=squid_height - (self.junction_width * 2),
                i=2,
                x=squid_width - x_offset - self.wire_width,
                y=y_position + self.junction_width,
            )
            shadow_shapes.insert(ws_right * squid_transform)

        # SQUID Debug Gap Boxes
        for i in range(int(squid_count - 1)):
            semi_wire_width = self.wire_width / 2
            semi_squid_width = squid_width / 2
            starting_x = semi_squid_width - semi_wire_width
            ending_x = starting_x + self.wire_width
            shape = pya.DBox(starting_x, 0, ending_x, gap_size)
            y_position = (i + 1) * squid_height + i * gap_size
            gap_transform = pya.DTrans(0, False, 0, y_position)
            junction_shapes.insert(gap_transform * shape * squid_transform)
            ws = wire_shadow(gap_size, i, starting_x, y_position)
            shadow_shapes.insert(ws * squid_transform)

        # CONNECTORS
        x_offset = self.squid_x_connector_offset - (self.wire_width / 2)
        raw_connector_length = (self.inductor_height - self.squid_area_height) / 2
        connector_length = raw_connector_length + self.finger_overlap
        wire_top = pya.DBox(x_offset, 0, x_offset + self.wire_width, connector_length)
        wire_top_transform = pya.DTrans(
            0,
            False,
            0,
            self.squid_area_height,
        )
        junction_shapes.insert(squid_transform * wire_top * wire_top_transform)
        wire_bottom = pya.DBox(x_offset, -connector_length, x_offset + self.wire_width, 0)
        wire_bottom_transform = pya.DTrans(
            0,
            False,
            0,
            0,
        )
        junction_shapes.insert(squid_transform * wire_bottom * wire_bottom_transform)
        ws_top = wire_shadow(raw_connector_length, 2, x_offset, 0)
        shadow_shapes.insert(ws_top * squid_transform * wire_top_transform)
        ws_bottom = wire_shadow(raw_connector_length, 2, x_offset, -raw_connector_length)
        shadow_shapes.insert(ws_bottom * squid_transform * wire_bottom_transform)

        # TAPPER North
        taper_unit_width = self.junction_length
        taper_unit_height = self.squid_wire_length
        double_connect_offset = self.squid_x_connector_offset * 2
        taper_north_offset_x = squid_offset_x + taper_unit_width - double_connect_offset
        taper_north_offset_y = squid_offset_y + self.squid_area_height
        taper_north_transform = pya.DTrans(0, False, taper_north_offset_x, taper_north_offset_y)
        taper_steps = int(self.taper_horizontal_displacement / self.taper_horizontal_displacement_step)
        adjusted_taper_horizontal_displacement = taper_steps * self.taper_horizontal_displacement_step
        last_taper_north_transform = None
        for i in range(taper_steps):
            taper_step_transform = pya.DTrans(
                0, False, i * self.taper_horizontal_displacement_step, i * taper_unit_height
            )
            # pylon
            if i == 0:
                x_offset = self.squid_x_connector_offset - (self.wire_width / 2)
            else:
                x_offset = (taper_unit_width / 2) - (self.wire_width / 2)
            pylon = pya.DBox(x_offset, 0, x_offset + self.wire_width, taper_unit_height)
            junction_shapes.insert(taper_step_transform * pylon * taper_north_transform)
            # junction
            junction = pya.DBox(0, taper_unit_height - self.junction_width, taper_unit_width, taper_unit_height)
            junction_shapes.insert(taper_step_transform * junction * taper_north_transform)
            # shadow
            ws = wire_shadow(taper_unit_height - self.junction_width, i - 1, x_offset, 0)
            shadow_shapes.insert(ws * taper_north_transform * taper_step_transform)
            [jsl, jsr] = junction_shadow(taper_unit_width, self.junction_width)
            shadow_transform = pya.DTrans(0, False, 0, taper_unit_height - self.junction_width)
            shadow_shapes.insert(jsl * taper_north_transform * taper_step_transform * shadow_transform)
            shadow_shapes.insert(jsr * taper_north_transform * taper_step_transform * shadow_transform)

        last_taper_north_step_count = taper_steps
        last_taper_north_transform = taper_north_transform * pya.DTrans(
            0,
            False,
            adjusted_taper_horizontal_displacement - taper_unit_width,
            last_taper_north_step_count * taper_unit_height,
        )

        # TAPPER South
        taper_unit_width = self.junction_length
        taper_unit_height = self.squid_wire_length
        double_connect_offset = self.squid_x_connector_offset * 2
        taper_south_offset_x = squid_offset_x + taper_unit_width - double_connect_offset
        taper_south_offset_y = squid_offset_y - self.squid_wire_length
        taper_south_transform = pya.DTrans(0, False, taper_south_offset_x, taper_south_offset_y)
        taper_steps = int(self.taper_horizontal_displacement / self.taper_horizontal_displacement_step)
        adjusted_taper_horizontal_displacement = taper_steps * self.taper_horizontal_displacement_step
        for i in range(taper_steps):
            taper_step_transform = pya.DTrans(
                0, False, i * self.taper_horizontal_displacement_step, -i * taper_unit_height
            )
            # pylon
            if i == 0:
                x_offset = self.squid_x_connector_offset - (self.wire_width / 2)
            else:
                x_offset = (taper_unit_width / 2) - (self.wire_width / 2)
            pylon = pya.DBox(x_offset, 0, x_offset + self.wire_width, taper_unit_height)
            junction_shapes.insert(taper_step_transform * pylon * taper_south_transform)
            # junction
            junction = pya.DBox(0, 0, taper_unit_width, self.junction_width)
            junction_shapes.insert(taper_step_transform * junction * taper_south_transform)
            # shadow
            ws = wire_shadow(taper_unit_height - self.junction_width, i - 1, x_offset, self.junction_width)
            shadow_shapes.insert(ws * taper_south_transform * taper_step_transform)
            [jsl, jsr] = junction_shadow(taper_unit_width, self.junction_width)
            shadow_shapes.insert(jsl * taper_south_transform * taper_step_transform)
            shadow_shapes.insert(jsr * taper_south_transform * taper_step_transform)

        last_taper_south_step_count = taper_steps
        last_taper_south_transform = taper_south_transform * pya.DTrans(
            0,
            False,
            adjusted_taper_horizontal_displacement - taper_unit_width,
            -(last_taper_north_step_count - 1) * taper_unit_height,
        )

        bend_height = self.squid_wire_length

        # Tower NordWest
        tower_unit_width = self.junction_length
        tower_unit_height = self.squid_wire_length
        tower_north_transform = last_taper_north_transform
        last_point = pya.DPoint(0, 0) * tower_north_transform
        available_height = (self.inductor_height - (self.inductor_padding + bend_height)) - last_point.y
        stored_available_height = available_height
        tower_desired_height = available_height
        tower_steps = int(tower_desired_height / tower_unit_height)
        adjusted_tower_height = tower_steps * tower_unit_height
        tower_unit_height = adjusted_tower_height / tower_steps
        for i in range(tower_steps):
            tower_step_transform = pya.DTrans(0, False, tower_unit_width, i * tower_unit_height)
            x_offset = (tower_unit_width / 2) - (self.wire_width / 2)
            # pylon
            pylon = pya.DBox(x_offset, 0, x_offset + self.wire_width, tower_unit_height)
            junction_shapes.insert(tower_step_transform * pylon * tower_north_transform)
            # junction
            junction = pya.DBox(0, tower_unit_height - self.junction_width, tower_unit_width, tower_unit_height)
            junction_shapes.insert(tower_step_transform * junction * tower_north_transform)
            # shadow
            ws = wire_shadow(tower_unit_height - self.junction_width, i + last_taper_north_step_count - 1, x_offset, 0)
            shadow_shapes.insert(ws * tower_north_transform * tower_step_transform)
            [jsl, jsr] = junction_shadow(tower_unit_width, self.junction_width)
            shadow_transform = pya.DTrans(0, False, 0, tower_unit_height - (self.junction_width))
            shadow_shapes.insert(jsl * tower_north_transform * tower_step_transform * shadow_transform)
            shadow_shapes.insert(jsr * tower_north_transform * tower_step_transform * shadow_transform)

        # North Bend
        last_tower_north_transform = (
            pya.DTrans(0, False, tower_unit_width, tower_steps * tower_unit_height) * tower_north_transform
        )
        bend_unit_width = self.junction_length * self.bend_width_multiplier
        bend_unit_height = self.squid_wire_length
        x_offset = self.squid_x_connector_offset - (self.wire_width / 2)
        bend_transform = pya.DTrans(0, False, tower_unit_width - (x_offset * 2), 0)
        pylon_left = pya.DBox(x_offset, 0, x_offset + self.wire_width, bend_unit_height)
        junction_shapes.insert(bend_transform * pylon_left * last_tower_north_transform)
        pylon_right = pya.DBox(
            bend_unit_width - x_offset, 0, bend_unit_width - x_offset + self.wire_width, bend_unit_height
        )
        junction_shapes.insert(bend_transform * pylon_right * last_tower_north_transform)
        bend_junction = pya.DBox(0, bend_unit_height - self.junction_width, bend_unit_width, bend_unit_height)
        junction_shapes.insert(bend_transform * bend_junction * last_tower_north_transform)
        # shadow
        wsl = wire_shadow(bend_unit_height - self.junction_width, 0, x_offset, 0)
        shadow_shapes.insert(wsl * last_tower_north_transform * bend_transform)
        wsr = wire_shadow(bend_unit_height - self.junction_width, 1, x_offset, 0)
        wsr_transform = pya.DTrans(0, False, bend_unit_width - (x_offset * 2), 0)
        shadow_shapes.insert(wsr * last_tower_north_transform * bend_transform * wsr_transform)
        [jsl, jsr] = junction_shadow(bend_unit_width, self.junction_width)
        js_transform = pya.DTrans(0, False, 0, bend_unit_height - self.junction_width)
        shadow_shapes.insert(jsl * last_tower_north_transform * bend_transform * js_transform)
        shadow_shapes.insert(jsr * last_tower_north_transform * bend_transform * js_transform)

        # Tower NordEast
        tower_unit_width = self.junction_length
        tower_unit_height = self.squid_wire_length
        x_offset = self.squid_x_connector_offset - (self.wire_width / 2)
        last_bend_transform = pya.DTrans(0, False, bend_unit_width - (x_offset * 2), 0)
        tower_north_bend_transform = last_tower_north_transform * last_bend_transform
        last_point = pya.DPoint(0, 0) * tower_north_bend_transform
        available_height = (self.inductor_height / 2) - (self.inductor_padding + bend_height)
        tower_north_east_transform = pya.DTrans(
            0, False, last_point.x - (x_offset * 2), last_point.y - available_height
        )
        tower_desired_height = available_height
        tower_steps = int(tower_desired_height / tower_unit_height)
        adjusted_tower_height = tower_steps * tower_unit_height
        tower_unit_height = adjusted_tower_height / tower_steps
        for i in range(tower_steps):
            x_offset = (tower_unit_width / 2) - (self.wire_width / 2)
            tower_step_transform = pya.DTrans(0, False, tower_unit_width, i * tower_unit_height)
            # wire
            pylon = pya.DBox(x_offset, 0, x_offset + self.wire_width, tower_unit_height)
            junction_shapes.insert(tower_step_transform * pylon * tower_north_east_transform)
            # junction
            junction = pya.DBox(0, tower_unit_height - self.junction_width, tower_unit_width, tower_unit_height)
            junction_shapes.insert(tower_step_transform * junction * tower_north_east_transform)
            # shadow
            last_shadow_offset = 0
            if i == 0:
                last_shadow_offset = -self.junction_width / 2
            ws = wire_shadow(
                tower_unit_height - self.junction_width + last_shadow_offset,
                i + last_taper_north_step_count - 1,
                x_offset,
                -last_shadow_offset,
            )
            shadow_shapes.insert(ws * tower_north_east_transform * tower_step_transform)
            [jsl, jsr] = junction_shadow(tower_unit_width, self.junction_width)
            shadow_transform = pya.DTrans(0, False, 0, tower_unit_height - (self.junction_width))
            shadow_shapes.insert(jsl * tower_north_east_transform * tower_step_transform * shadow_transform)
            shadow_shapes.insert(jsr * tower_north_east_transform * tower_step_transform * shadow_transform)

        # Tower SouthWest (Inverted version of Tower NordWest)
        tower_unit_width = self.junction_length
        tower_unit_height = self.squid_wire_length
        tower_south_transform = last_taper_south_transform
        last_point = pya.DPoint(0, 0) * tower_south_transform
        available_height = (
            stored_available_height  # ((self.inductor_height) - (self.inductor_padding + bend_height)) - last_point.y
        )
        tower_desired_height = available_height
        tower_steps = int(tower_desired_height / tower_unit_height)
        adjusted_tower_height = tower_steps * tower_unit_height
        tower_unit_height = adjusted_tower_height / tower_steps
        for i in range(tower_steps):
            tower_step_transform = pya.DTrans(0, False, tower_unit_width, -i * tower_unit_height)
            x_offset = (tower_unit_width / 2) - (self.wire_width / 2)
            # wire
            pylon = pya.DBox(x_offset, -tower_unit_height, x_offset + self.wire_width, 0)
            junction_shapes.insert(tower_step_transform * pylon * tower_south_transform)
            # junction
            junction = pya.DBox(0, -tower_unit_height, tower_unit_width, -tower_unit_height + self.junction_width)
            junction_shapes.insert(tower_step_transform * junction * tower_south_transform)
            # shadow
            ws = wire_shadow(
                tower_unit_height - self.junction_width,
                i + last_taper_south_step_count - 1,
                x_offset,
                self.junction_width,
            )
            shadow_transform = pya.DTrans(0, False, 0, -tower_unit_height)
            shadow_shapes.insert(ws * tower_south_transform * tower_step_transform * shadow_transform)
            [jsl, jsr] = junction_shadow(tower_unit_width, self.junction_width)
            shadow_shapes.insert(jsl * tower_south_transform * tower_step_transform * shadow_transform)
            shadow_shapes.insert(jsr * tower_south_transform * tower_step_transform * shadow_transform)

        # South Bend (Inverted version of North Bend)
        last_tower_south_transform = (
            pya.DTrans(0, False, tower_unit_width, -tower_steps * tower_unit_height) * tower_south_transform
        )
        bend_unit_width = self.junction_length * self.bend_width_multiplier
        bend_unit_height = self.squid_wire_length
        x_offset = self.squid_x_connector_offset - (self.wire_width / 2)
        bend_transform = pya.DTrans(0, False, tower_unit_width - (x_offset * 2), 0)
        pylon_left = pya.DBox(x_offset, -bend_unit_height, x_offset + self.wire_width, 0)
        junction_shapes.insert(bend_transform * pylon_left * last_tower_south_transform)
        pylon_right = pya.DBox(
            bend_unit_width - x_offset, -bend_unit_height, bend_unit_width - x_offset + self.wire_width, 0
        )
        junction_shapes.insert(bend_transform * pylon_right * last_tower_south_transform)
        bend_junction = pya.DBox(0, -bend_unit_height, bend_unit_width, -bend_unit_height + self.junction_width)
        junction_shapes.insert(bend_transform * bend_junction * last_tower_south_transform)
        # shadow
        shadow_bend_transform = pya.DTrans(0, False, 0, -bend_unit_height)
        wsl = wire_shadow(bend_unit_height - self.junction_width, 0, x_offset, self.junction_width)
        shadow_shapes.insert(wsl * last_tower_south_transform * bend_transform * shadow_bend_transform)
        wsr = wire_shadow(bend_unit_height - self.junction_width, 1, x_offset, self.junction_width)
        wsr_transform = pya.DTrans(0, False, bend_unit_width - (x_offset * 2), 0)
        shadow_shapes.insert(wsr * last_tower_south_transform * bend_transform * wsr_transform * shadow_bend_transform)
        [jsl, jsr] = junction_shadow(bend_unit_width, self.junction_width)
        shadow_shapes.insert(jsl * last_tower_south_transform * bend_transform * shadow_bend_transform)
        shadow_shapes.insert(jsr * last_tower_south_transform * bend_transform * shadow_bend_transform)

        # Tower SouthEast (Inverted version of Tower NordEast)
        tower_unit_width = self.junction_length
        tower_unit_height = self.squid_wire_length
        x_offset = self.squid_x_connector_offset - (self.wire_width / 2)
        last_bend_transform = pya.DTrans(0, False, bend_unit_width - (x_offset * 2), 0)
        tower_south_bend_transform = last_tower_south_transform * last_bend_transform
        last_point = pya.DPoint(0, 0) * tower_south_bend_transform
        available_height = (self.inductor_height / 2) - (self.inductor_padding + bend_height)
        tower_south_east_transform = pya.DTrans(
            0, False, last_point.x - (x_offset * 2), last_point.y + available_height
        )
        tower_desired_height = available_height
        tower_steps = int(tower_desired_height / tower_unit_height)
        adjusted_tower_height = tower_steps * tower_unit_height
        tower_unit_height = adjusted_tower_height / tower_steps
        last_mid_point = None
        for i in range(tower_steps):
            x_offset = (tower_unit_width / 2) - (self.wire_width / 2)
            tower_step_transform = pya.DTrans(0, False, tower_unit_width, -i * tower_unit_height)
            # wire
            pylon = pya.DBox(x_offset, -tower_unit_height, x_offset + self.wire_width, 0)
            junction_shapes.insert(tower_step_transform * pylon * tower_south_east_transform)
            last_mid_point = tower_step_transform * tower_south_east_transform
            # junction
            junction = pya.DBox(0, -tower_unit_height, tower_unit_width, -tower_unit_height + self.junction_width)
            junction_shapes.insert(tower_step_transform * junction * tower_south_east_transform)
            # shadow
            last_shadow_offset = 0
            if i == 0:
                last_shadow_offset = -self.junction_width / 2
            shadow_transform = pya.DTrans(0, False, 0, -tower_unit_height)
            ws = wire_shadow(
                tower_unit_height - self.junction_width + last_shadow_offset,
                i + last_taper_north_step_count - 1,
                x_offset,
                self.junction_width,
            )
            shadow_shapes.insert(ws * tower_south_east_transform * tower_step_transform * shadow_transform)
            [jsl, jsr] = junction_shadow(tower_unit_width, self.junction_width)
            shadow_shapes.insert(jsl * tower_south_east_transform * tower_step_transform * shadow_transform)
            shadow_shapes.insert(jsr * tower_south_east_transform * tower_step_transform * shadow_transform)

        # Phase Slip Junction
        slip_unit_width = self.phase_slip_junction_length
        slip_unit_height = self.junction_width
        last_tower_point = pya.DPoint(0, 0) * last_mid_point
        phase_junction = pya.DBox(
            -slip_unit_width / 2, -slip_unit_height / 2, slip_unit_width / 2, slip_unit_height / 2
        )
        phase_junction_transform = pya.DTrans(
            0, False, last_tower_point.x + (tower_unit_width / 2), self.inductor_height / 2
        )
        junction_shapes.insert(phase_junction * phase_junction_transform)
        phase_junction_shadow_transform = pya.DTrans(
            0,
            False,
            last_tower_point.x + ((tower_unit_width - slip_unit_width) / 2),
            (self.inductor_height / 2) - (self.junction_width / 2),
        )
        [jsl, jsr] = junction_shadow(slip_unit_width, self.junction_width)
        shadow_shapes.insert(jsl * phase_junction_shadow_transform)
        shadow_shapes.insert(jsr * phase_junction_shadow_transform)

    def _add_shapes(self, shapes, layer):
        """Merge shapes into a region and add it to layer."""
        region = pya.Region(shapes).merged()
        self.cell.shapes(self.get_layer(layer)).insert(region)

    def _add_refpoints(self):
        """Adds the "origin_squid" refpoint and port "common"."""
        self.refpoints["origin_squid"] = pya.DPoint(0, 0)
        self.add_port("common", pya.DPoint(0, self.inductor_height))

    def _produce_ground_metal_shapes(self):
        """Produces hardcoded shapes in metal gap and metal addition layers."""
        # metal additions bottom
        # x0 = -self.a / 2
        # y0 = self.inductor_height / 2
        bottom_pts = []
        if self.include_base_metal_addition:
            shape = polygon_with_vsym(bottom_pts)
            # self.cell.shapes(self.get_layer("base_metal_addition")).insert(shape)
            # metal additions top
            top_pts = []
            shape = polygon_with_vsym(top_pts)
            self.cell.shapes(self.get_layer("base_metal_addition")).insert(shape)
        # metal gap
        if self.include_base_metal_gap:
            if self.include_base_metal_addition:
                pts = (
                    bottom_pts
                    + [
                        pya.DPoint(-self.inductor_width / 2, 0),
                        pya.DPoint(-self.inductor_width / 2, self.inductor_height),
                    ]
                    + top_pts[::-1]
                )
            else:
                pts = [
                    pya.DPoint(-self.inductor_width / 2, 0),
                    pya.DPoint(-self.inductor_width / 2, self.inductor_height),
                ]
            shape = polygon_with_vsym(pts)
            self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(shape)

    def _produce_ground_grid_avoidance(self):
        """Add ground grid avoidance."""
        w = self.cell.dbbox().width()
        h = self.cell.dbbox().height()
        protection = pya.DBox(-w / 2 - self.margin, -self.margin, w / 2 + self.margin, h + self.margin)
        self.add_protection(protection)

    def _round_corners_and_append(self, polygon, polygon_list, rounding_params):
        """Rounds the corners of the polygon, converts it to integer coordinates, and adds it to the polygon list."""
        polygon = polygon.round_corners(rounding_params["rinner"], rounding_params["router"], rounding_params["n"])
        polygon_list.append(polygon.to_itype(self.layout.dbu))



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).


import os
import stat
import logging
import json
import argparse
import platform
import copy

from dataclasses import replace
from pathlib import Path
from typing import Sequence, Union, Tuple, Dict, Optional

from kqcircuits.simulations.export.simulation_export import (
    copy_content_into_directory,
    get_post_process_command_lines,
    get_combined_parameters,
    export_simulation_json,
)
from kqcircuits.simulations.export.simulation_validate import validate_simulation
from kqcircuits.util.load_save_layout import save_layout
from kqcircuits.util.export_helper import write_commit_reference_file
from kqcircuits.defaults import ELMER_SCRIPT_PATHS, KQC_REMOTE_ACCOUNT, SIM_SCRIPT_PATH
from kqcircuits.simulations.simulation import Simulation
from kqcircuits.simulations.cross_section_simulation import CrossSectionSimulation
from kqcircuits.simulations.export.elmer.elmer_solution import ElmerEPR3DSolution, ElmerSolution, get_elmer_solution
from kqcircuits.simulations.post_process import PostProcess


def export_elmer_json(
    simulation: Union[Simulation, CrossSectionSimulation],
    solution: ElmerSolution,
    path: Path,
    workflow: dict,
    mesh_reuse_name: str | None = None,
):
    """
    Export Elmer simulation into json and gds files.

    Args:
        simulation: The simulation to be exported.
        solution: The solution to be exported.
        path: Location where to write json and gds files.
        workflow: Parameters for simulation workflow
        mesh_reuse_name: Name of a mesh to be reused from another simulation

    Returns:
         Path to exported json file.
    """
    is_cross_section = isinstance(simulation, CrossSectionSimulation)

    if simulation is None or not isinstance(simulation, (Simulation, CrossSectionSimulation)):
        raise ValueError("Cannot export without simulation")

    # write .gds file
    gds_file = simulation.name + ".gds"
    gds_file_path = str(path.joinpath(gds_file))
    if not Path(gds_file_path).exists():
        save_layout(gds_file_path, simulation.layout, [simulation.cell], simulation.get_layers(), no_empty_cells=True)

    sim_data = simulation.get_simulation_data()
    sol_data = solution.get_solution_data()
    full_name = simulation.name + solution.name

    if is_cross_section:
        sif_names = [f"{full_name}_C"]
        if sol_data["run_inductance_sim"]:
            if any(m.get("london_penetration_depth", 0) > 0 for m in sim_data["material_dict"].values()):
                sif_names += [f"{full_name}_L"]
            else:
                sif_names += [f"{full_name}_C0"]
    elif solution.tool == "wave_equation":
        if solution.sweep_type == "interpolating":
            sif_names = []
        else:
            sif_names = [full_name + "_f" + str(f).replace(".", "_") for f in sol_data["frequency"]]
    else:
        sif_names = [full_name]

    json_data = {
        "name": full_name,
        "mesh_name": mesh_reuse_name if mesh_reuse_name else full_name,
        "workflow": workflow,
        **sim_data,
        **sol_data,
        "sif_names": sif_names,
        "gds_file": gds_file,
        "parameters": get_combined_parameters(simulation, solution),
    }

    # write .json file
    json_file_path = str(path.joinpath(full_name + ".json"))
    export_simulation_json(json_data, json_file_path)

    return json_file_path


def export_elmer_script(
    json_filenames,
    path: Path,
    workflow=None,
    script_folder: str = "scripts",
    file_prefix: str = "simulation",
    script_file: str = "run.py",
    post_process=None,
    compile_elmer_modules=False,
):
    """
    Create script files for running one or more simulations.
    Create also a main script to launch all the simulations at once.

    Args:
        json_filenames: List of paths to json files to be included into the script.
        path: Location where to write the script file.
        workflow: Parameters for simulation workflow
        script_folder: Path to the Elmer-scripts folder.
        file_prefix: File prefix of the script file to be created.
        script_file: Name of the script file to run.
        post_process: List of PostProcess objects, a single PostProcess object, or None to be executed after simulations
        compile_elmer_modules: Compile custom Elmer energy integration module at runtime.

    Returns:

        Path of exported main script file
    """

    if workflow is None:
        workflow = {}
    sbatch = "sbatch_parameters" in workflow

    python_executable = workflow.get("python_executable", "python")
    main_script_filename = str(path.joinpath(file_prefix + ".sh"))
    execution_script = Path(script_folder).joinpath(script_file)

    n_jsons = len(json_filenames)

    elmer_compile_str = 'echo "Compiling Elmer modules"\n'

    if platform.system() == "Windows":
        elmer_compile_str += (
            f'cmd //c "elmerf90 -fcheck=all {script_folder}/SaveBoundaryEnergy.F90 -o SaveBoundaryEnergy.dll"\n'
        )
    else:
        elmer_compile_str += (
            f"elmerf90 -fcheck=all {script_folder}/SaveBoundaryEnergy.F90 -o SaveBoundaryEnergy > /dev/null\n"
        )

    path.joinpath("log_files").mkdir(parents=True, exist_ok=True)

    def _write_script(filename, lines, interp_line="#!/bin/bash\n"):
        """Writes a script and makes it executable.
        `lines` is a string or list of strings that will be written into `filename`"""
        with open(filename, "w", encoding="utf-8") as file:
            if isinstance(lines, list):
                lines = "".join(lines)
            file.write(interp_line + lines)
        # change permissions
        os.chmod(filename, os.stat(filename).st_mode | stat.S_IEXEC)

    def _get_from_json(json_filename, keys):
        with open(json_filename, encoding="utf-8") as f:
            json_data = json.load(f)
            # return as a tuple
            return (*[json_data[k] for k in keys],)

    def _sim_part_echo(i, part):
        return f'echo "Simulation {i + 1}/{n_jsons} {part}"\n'

    if sbatch:

        def _multiply_time(time_str, multiplier):
            """
            Helper function to multiply a time of format "HH:MM:SS" with a constant. In
            this case, we multiply timeout per simulation by the number of simulations

            Args:
                time_str (str): Time in format "HH:MM:SS"
                multiplier (float): multiplier

            Returns:
                New time string in format "HH:MM:SS"
            """
            time_str = time_str.strip()
            if len(time_str) != 8:
                raise ValueError('Invalid sbatch/slurm time formatting! Format has to be "HH:MM:SS"')
            hours = int(int(time_str[0:2]) * multiplier)
            minutes = int(int(time_str[3:5]) * multiplier)
            seconds = int(int(time_str[6:8]) * multiplier)
            if seconds > 60:
                minutes = minutes + seconds // 60
                seconds = seconds % 60
            if minutes > 60:
                hours = hours + minutes // 60
                minutes = minutes % 60
            return f"{hours:02d}:{minutes:02d}:{seconds:02d}"

        def _multiply_mem(mem_str, multiplier):
            """
            Helper function to multiply a memory specification with a constant. If multiplier < 1,
            will convert to using smaller units if possible

            Args:
                mem_str (str): Amount of memory in format "1234X" where "X" specifies the unit
                               and anything before is a number specifying the amount
                multiplier (float): multiplier

            Returns:
                New memory string
            """
            original_memstr = mem_str.strip()
            unit = original_memstr[-1]
            if unit in ("K", "M", "G", "T"):
                original_mem_int = int(original_memstr.partition(unit)[0])
            else:
                original_mem_int = int(original_memstr)
                unit = "M"
            downconversion = {"M": "K", "G": "M", "T": "G"}
            if multiplier < 1.0 and unit != "K":
                elmer_mem_per_f = int(original_mem_int * 1024 * multiplier)
                elmer_mem_per_f = str(elmer_mem_per_f) + downconversion[unit]
            else:
                elmer_mem_per_f = int(original_mem_int * multiplier)
                elmer_mem_per_f = str(elmer_mem_per_f) + unit
            return elmer_mem_per_f

        def _divup(a, b):
            return -(a // -b)

        def _get_srun_command(nodes, tasks, cores, memory):
            return f"srun -N {nodes} -n {tasks} -c {cores} --cpu-bind none --exact --mem={memory}"

        def _get_sbatch_lines(sbatch_settings):
            lines = []
            for s_key, s_value in sbatch_settings.items():
                lines.append(f"#SBATCH {s_key}={s_value}\n")
            lines += [
                "\n# set the number of threads based on --cpus-per-task\n",
                "export OMP_NUM_THREADS=$SLURM_CPUS_PER_TASK\n\n",
            ]
            return lines

        sbatch_parameters = workflow["sbatch_parameters"]

        parallelization_level = workflow["_parallelization_level"]
        n_simulations = workflow["_n_simulations"]
        _n_workers = int(sbatch_parameters.pop("n_workers", 1))

        if parallelization_level == "full_simulation":
            n_workers_full = _n_workers
            n_workers_elmer_only = 1
            n_simulations_gmsh = n_simulations
        elif parallelization_level == "elmer":
            n_workers_full = 1
            n_workers_elmer_only = _n_workers
            n_simulations_gmsh = 1
        elif parallelization_level == "none":
            n_workers_full = 1
            n_workers_elmer_only = 1
            n_simulations_gmsh = 1
        else:
            raise ValueError(f"Unknown parallelization level {parallelization_level}")

        if sbatch_parameters.get("--account", "project_0") == "project_0":
            sbatch_parameters["--account"] = KQC_REMOTE_ACCOUNT

        common_keys = [k for k in sbatch_parameters.keys() if k.startswith("--")]
        sbatch_settings_elmer = {k: sbatch_parameters.pop(k) for k in common_keys}

        sbatch_settings_meshes = sbatch_settings_elmer.copy()

        max_cpus_per_node = int(sbatch_parameters.pop("max_threads_per_node", 40))

        elmer_tasks_per_worker = int(sbatch_parameters.pop("elmer_n_processes", 10))
        elmer_cpus_per_task = int(sbatch_parameters.pop("elmer_n_threads", 1))
        if elmer_cpus_per_task > 1 and elmer_tasks_per_worker > 1:
            logging.warning(
                "Using both process and thread level parallelization with Elmer might result in poor performance"
            )
        elmer_cpus_per_worker = elmer_tasks_per_worker * elmer_cpus_per_task

        elmer_mem_per_worker = sbatch_parameters.pop("elmer_mem", "64G")

        if elmer_cpus_per_worker > max_cpus_per_node:
            elmer_nodes_per_worker = _divup(elmer_cpus_per_worker, max_cpus_per_node)
            elmer_total_nodes = elmer_nodes_per_worker * _n_workers
            elmer_tasks_per_node = min(elmer_tasks_per_worker, max_cpus_per_node)
            elmer_mem_per_node = _multiply_mem(elmer_mem_per_worker, _n_workers / elmer_total_nodes)
        else:
            elmer_nodes_per_worker = 1
            elmer_max_workers_per_node = max_cpus_per_node // elmer_cpus_per_worker
            elmer_total_nodes = _divup(_n_workers, elmer_max_workers_per_node)
            elmer_workers_per_node = min(_n_workers, elmer_max_workers_per_node)
            elmer_tasks_per_node = elmer_workers_per_node * elmer_tasks_per_worker
            elmer_mem_per_node = _multiply_mem(elmer_mem_per_worker, elmer_workers_per_node)

        gmsh_cpus_per_worker = int(sbatch_parameters.pop("gmsh_n_threads", 10))
        if gmsh_cpus_per_worker > max_cpus_per_node:
            raise RuntimeError(
                f"Requested more gmsh threads per worker {gmsh_cpus_per_worker}"
                f" than the limit per node {max_cpus_per_node}"
            )
        gmsh_max_workers_per_node = max_cpus_per_node // gmsh_cpus_per_worker
        gmsh_n_nodes = _divup(n_workers_full, gmsh_max_workers_per_node)
        gmsh_mem_per_worker = sbatch_parameters.pop("gmsh_mem", "64G")

        gmsh_workers_per_node = min(n_workers_full, gmsh_max_workers_per_node)
        gmsh_mem_per_node = _multiply_mem(gmsh_mem_per_worker, gmsh_workers_per_node)

        sbatch_settings_elmer["--time"] = _multiply_time(
            sbatch_parameters.pop("elmer_time", "00:10:00"), _divup(n_simulations, _n_workers)
        )
        sbatch_settings_elmer["--nodes"] = elmer_total_nodes
        sbatch_settings_elmer["--ntasks-per-node"] = elmer_tasks_per_node
        sbatch_settings_elmer["--cpus-per-task"] = elmer_cpus_per_task
        sbatch_settings_elmer["--mem"] = elmer_mem_per_node

        sbatch_settings_meshes["--time"] = _multiply_time(
            sbatch_parameters.pop("gmsh_time", "00:10:00"), _divup(n_simulations_gmsh, n_workers_full)
        )
        sbatch_settings_meshes["--nodes"] = gmsh_n_nodes
        sbatch_settings_meshes["--ntasks-per-node"] = gmsh_workers_per_node
        sbatch_settings_meshes["--cpus-per-task"] = gmsh_cpus_per_worker
        sbatch_settings_meshes["--mem"] = gmsh_mem_per_node

        srun_cmd_gmsh = _get_srun_command(1, 1, gmsh_cpus_per_worker, gmsh_mem_per_worker)
        srun_cmd_script = _get_srun_command(1, 1, 1, elmer_mem_per_worker)
        srun_cmd_elmer = _get_srun_command(
            elmer_nodes_per_worker, elmer_tasks_per_worker, elmer_cpus_per_task, elmer_mem_per_worker
        )

        env_setup = [s.removesuffix("\n") + "\n" for s in sbatch_parameters.pop("env_setup_cmds", [])]

        if len(sbatch_parameters) > 0:
            logging.warning("Unused sbatch parameters: ")
            for k, v in sbatch_parameters.items():
                logging.warning(f"{k} : {v}")

        dep_mesh_scripts = []
        indep_mesh_scripts = []
        for i, json_filename in enumerate(json_filenames):

            (simulation_name, mesh_name) = _get_from_json(json_filename, ["name", "mesh_name"])
            python_run_cmd = f'{python_executable} -u "{execution_script}" "{Path(json_filename).relative_to(path)}"'

            def get_log_cmd(logfile_suffix, filename=simulation_name):
                return f'2>&1 >> "log_files/{filename}.{logfile_suffix}.log"\n'

            script_lines = ["set -e\n"]

            if mesh_name == simulation_name:
                script_lines += [
                    _sim_part_echo(i, "Gmsh"),
                    f'{srun_cmd_gmsh} {python_run_cmd} --only-gmsh -q {get_log_cmd("Gmsh")}',
                    _sim_part_echo(i, "ElmerGrid"),
                    f'{srun_cmd_gmsh} ElmerGrid 14 2 "{simulation_name}.msh" {get_log_cmd("ElmerGrid")}',
                ]

                if int(elmer_tasks_per_worker) > 1:
                    script_lines.append(
                        f'{srun_cmd_gmsh} ElmerGrid 2 2 "{simulation_name}" -metis {elmer_tasks_per_worker}'
                        f' 4 --partdual --removeunused {get_log_cmd("ElmerGrid")}'
                    )

            script_lines += [
                _sim_part_echo(i, "Write Elmer sif files"),
                f'{srun_cmd_gmsh} {python_run_cmd} --only-elmer-sifs {get_log_cmd("Elmer_sifs")}',
            ]

            script_filename_meshes = str(path.joinpath(simulation_name + "_meshes.sh"))
            _write_script(script_filename_meshes, script_lines)

            run_str = f'source "{Path(script_filename_meshes).relative_to(path)}" &\n'
            (indep_mesh_scripts if mesh_name == simulation_name else dep_mesh_scripts).append(run_str)

        meshes_script_lines = _get_sbatch_lines(sbatch_settings_meshes)
        meshes_script_lines += env_setup
        if compile_elmer_modules:
            meshes_script_lines.append(elmer_compile_str)

        meshes_script_lines += [
            s + ("wait\n" if i % n_workers_full == 0 or i == len(indep_mesh_scripts) else "")
            for i, s in enumerate(indep_mesh_scripts, 1)
        ]
        meshes_script_lines += [
            s + ("wait\n" if i % n_workers_full == 0 or i == len(dep_mesh_scripts) else "")
            for i, s in enumerate(dep_mesh_scripts, 1)
        ]

        _write_script(str(path.joinpath(file_prefix + "_meshes.sh")), meshes_script_lines)

        main_script_lines = _get_sbatch_lines(sbatch_settings_elmer)
        main_script_lines += env_setup

        for i, json_filename in enumerate(json_filenames):
            simulation_name, sif_names = _get_from_json(json_filename, ["name", "sif_names"])

            sifs_split = [
                sif_names[i : min(i + n_workers_elmer_only, len(sif_names))]
                for i in range(0, len(sif_names), n_workers_elmer_only)
            ]

            python_run_cmd = f'{python_executable} -u "{execution_script}" "{Path(json_filename).relative_to(path)}"'

            def get_log_cmd(logfile_suffix, filename=simulation_name):  # pylint: disable=function-redefined
                return f'2>&1 >> "log_files/{filename}.{logfile_suffix}.log"\n'

            script_lines = ["set -e\n", _sim_part_echo(i, "Elmer")]

            for sif_list in sifs_split:
                for sif in sif_list:
                    sif_path = f"{simulation_name}/{sif}.sif"
                    script_lines.append(
                        f'{srun_cmd_elmer} ElmerSolver_mpi "{sif_path}" 2>&1 >> "log_files/{sif}.Elmer.log" & \n'
                    )
                script_lines.append("wait\n")

            script_lines += [
                _sim_part_echo(i, "write results json"),
                f'{srun_cmd_script} {python_run_cmd} --write-project-results {get_log_cmd("write_project_results")}',
            ]

            script_filename = str(path.joinpath(simulation_name + ".sh"))
            _write_script(script_filename, script_lines)

            main_script_lines += [
                f'source "{Path(script_filename).relative_to(path)}" &\n',
            ]
            if (i + 1) % n_workers_full == 0 or (i + 1) == n_jsons:
                main_script_lines.append("wait\n")

        main_script_lines += [
            'echo "--------------------------------------------"\n',
            'echo "Write versions file"\n',
            f"{srun_cmd_script} {python_run_cmd} --write-versions-file\n",
        ]

    else:  # local workflow
        n_workers = workflow.get("n_workers", 1)
        parallelization_level = workflow["_parallelization_level"]
        parallelize_workload = parallelization_level == "full_simulation" and n_workers > 1

        main_script_lines = []

        if compile_elmer_modules:
            main_script_lines.append(elmer_compile_str)

        if parallelize_workload:
            main_script_lines.append(f"export OMP_NUM_THREADS={workflow['elmer_n_threads']}\n")
            main_script_lines.append(f"{python_executable} {script_folder}/simple_workload_manager.py {n_workers}")

        dependent_sims = []
        for i, json_filename in enumerate(json_filenames):
            (simulation_name, mesh_name) = _get_from_json(json_filename, ["name", "mesh_name"])
            python_run_cmd = f'{python_executable} "{execution_script}" "{Path(json_filename).relative_to(path)}"'

            def get_log_cmd(logfile_suffix, filename=simulation_name):
                return f'2>&1 >> "log_files/{filename}.{logfile_suffix}.log"\n'

            script_filename = str(path.joinpath(simulation_name + ".sh"))

            script_lines = [
                "set -e\n",
                _sim_part_echo(i, "Gmsh"),
                f'{python_run_cmd} --only-gmsh {get_log_cmd("Gmsh")}',
                _sim_part_echo(i, "ElmerGrid"),
                f'{python_run_cmd} --only-elmergrid {get_log_cmd("ElmerGrid")}',
                _sim_part_echo(i, "Write Elmer sif files"),
                f'{python_run_cmd} --only-elmer-sifs {get_log_cmd("Elmer_sifs")}',
                _sim_part_echo(i, "Elmer"),
                f"{python_run_cmd} --only-elmer\n",
                _sim_part_echo(i, "Paraview"),
                f"{python_run_cmd} --only-paraview\n",
                _sim_part_echo(i, "Write results json"),
                f'{python_run_cmd} --write-project-results {get_log_cmd("write_project_results")}',
            ]

            _write_script(script_filename, script_lines)

            if parallelize_workload:
                run_str = f' "./{Path(script_filename).relative_to(path)}"'
                if mesh_name == simulation_name:
                    main_script_lines.append(run_str)
                else:
                    dependent_sims.append(run_str)
            else:
                main_script_lines += [
                    f'echo "Submitting the main script of simulation {i + 1}/{n_jsons}"\n',
                    'echo "--------------------------------------------"\n',
                    f'"./{Path(script_filename).relative_to(path)}"\n',
                ]

        if dependent_sims:
            main_script_lines.append(f"\n{python_executable} {script_folder}/simple_workload_manager.py {n_workers}")
            main_script_lines += dependent_sims

        main_script_lines += [
            '\necho "--------------------------------------------"\n',
            'echo "Write versions file"\n',
            f"{python_run_cmd} --write-versions-file\n",
        ]

    main_script_lines.append("\n" + get_post_process_command_lines(post_process, path, json_filenames))
    _write_script(main_script_filename, main_script_lines)

    return main_script_filename


def export_elmer(
    simulations: Sequence[
        Union[
            Simulation,
            Tuple[Simulation, ElmerSolution],
            CrossSectionSimulation,
            Tuple[CrossSectionSimulation, ElmerSolution],
        ]
    ],
    path: Path,
    script_folder: str = "scripts",
    file_prefix: str = "simulation",
    script_file: str = "run.py",
    workflow: Optional[Dict] = None,
    skip_errors: bool = False,
    post_process: Optional[Union[PostProcess, Sequence[PostProcess]]] = None,
    **solution_params,
) -> Path:
    """
    Exports an elmer simulation model to the simulation path.

    Args:
        simulations: List of Simulation objects or tuples containing Simulation and Solution objects.
        path: Location where to output the simulation model
        script_folder: Path to the Elmer-scripts folder.
        file_prefix: File prefix of the script file to be created.
        script_file: Name of the script file to run.
        workflow: Parameters for simulation workflow
        skip_errors: Skip simulations that cause errors. (Default: False)

            .. warning::

               **Use this carefully**, some of your simulations might not make sense physically and
               you might end up wasting time on bad simulations.
        post_process: List of PostProcess objects, a single PostProcess object, or None to be executed after simulations
        solution_params: ElmerSolution parameters if simulations is a list of Simulation objects.

    Returns:
        Path to exported script file.
    """

    common_sol = None if all(isinstance(s, Sequence) for s in simulations) else get_elmer_solution(**solution_params)

    workflow = _update_elmer_workflow(simulations, common_sol, workflow)

    # If doing 3D epr simulations the custom Elmer energy integration module is compiled at runtime
    epr_sim = _is_epr_sim(simulations, common_sol)
    script_paths = ELMER_SCRIPT_PATHS + [SIM_SCRIPT_PATH / "elmer_modules"] if epr_sim else ELMER_SCRIPT_PATHS

    write_commit_reference_file(path)
    copy_content_into_directory(script_paths, path, script_folder)

    def make_names_elmer_compatible(sim, sol):
        """Replace dots with dashes and make lowercase"""
        sim.name = sim.name.replace(".", "-").lower()
        sol_name = sol.name.replace(".", "-").lower()
        if 2 * (len(sim.name) + len(sol_name)) + len("_20np01_t0001.vtu") >= 128:
            logging.warning(
                "Simulation and solution names might be too long for Elmer."
                "Try to shorten them or create the sweep manually"
            )

        return (sim, replace(sol, name=sol_name))

    simulations = [
        make_names_elmer_compatible(*(sim_sol if isinstance(sim_sol, Sequence) else (sim_sol, common_sol)))
        for sim_sol in simulations
    ]
    sim_objects, sol_objects = list(zip(*simulations))

    n_sim = len(simulations)
    mesh_reuse_name = n_sim * [None]
    for i in range(n_sim):
        for j in range(i):
            if (
                sim_objects[i] == sim_objects[j]
                and sol_objects[i].mesh_size == sol_objects[j].mesh_size
                and sol_objects[i].mesh_optimizer == sol_objects[j].mesh_optimizer
                and sol_objects[i].tool == sol_objects[j].tool
            ):
                mesh_reuse_name[i] = (
                    mesh_reuse_name[j] if mesh_reuse_name[j] else (sim_objects[j].name + sol_objects[j].name)
                )
                break

    json_filenames = []
    for simulation, solution, mesh_reuse in zip(sim_objects, sol_objects, mesh_reuse_name):
        validate_simulation(simulation, solution)

        try:
            json_filenames.append(export_elmer_json(simulation, solution, path, workflow, mesh_reuse))
        except (IndexError, ValueError, Exception) as e:  # pylint: disable=broad-except
            if skip_errors:
                logging.warning(
                    f"Simulation {simulation.name} skipped due to {e.args}. "
                    "Some of your other simulations might not make sense geometrically. "
                    "Disable `skip_errors` to see the full traceback."
                )
            else:
                raise UserWarning(
                    "Generating simulation failed. You can discard the errors using `skip_errors` in `export_elmer`. "
                    "Moreover, `skip_errors` enables visual inspection of failed and successful simulation "
                    "geometry files."
                ) from e

    return export_elmer_script(
        json_filenames,
        path,
        workflow,
        script_folder=script_folder,
        file_prefix=file_prefix,
        script_file=script_file,
        post_process=post_process,
        compile_elmer_modules=epr_sim,
    )


def _is_epr_sim(simulations, common_sol):
    """Helper to check if doing 3D epr simulation"""
    epr_sim = False
    if common_sol is None:
        if any(isinstance(simsol[1], ElmerEPR3DSolution) for simsol in simulations):
            epr_sim = True
    elif isinstance(common_sol, ElmerEPR3DSolution):
        epr_sim = True

    return epr_sim


def _update_elmer_workflow(simulations, common_solution, workflow):
    """
    Modify workflow based on number of simulations and available computing resources

    Args:
        simulations: List of Simulation objects or tuples containing Simulation and Solution objects.
        common_solution: Solution object if not contained in `simulations`
        workflow: workflow to be updated

    Returns:
        Updated workflow
    """
    if workflow is None:
        workflow = {}
    workflow = copy.deepcopy(workflow)
    parallelization_level = "none"
    n_worker_lim = 1
    num_sims = len(simulations)

    if num_sims == 1:
        sol_obj = simulations[0][1] if common_solution is None else common_solution

        if sol_obj.tool == "wave_equation" and len(sol_obj.frequency) > 1:
            parallelization_level = "elmer"
            n_worker_lim = len(sol_obj.frequency)
    elif num_sims > 1:
        # TODO enable Elmer level parallelism with solution sweep
        n_worker_lim = num_sims
        parallelization_level = "full_simulation"

    workflow["_parallelization_level"] = parallelization_level
    workflow["_n_simulations"] = n_worker_lim

    if "sbatch_parameters" in workflow:
        n_workers = workflow["sbatch_parameters"].get("n_workers", 1.0)
        workflow["sbatch_parameters"]["n_workers"] = min(int(n_workers), n_worker_lim)
        workflow.pop("elmer_n_processes", "")
        workflow.pop("elmer_n_threads", "")
        workflow.pop("n_workers", "")
        workflow.pop("gmsh_n_threads", "")
    else:

        n_workers = workflow.get("n_workers", 1)
        n_processes = workflow.get("elmer_n_processes", 1)
        n_threads = workflow.get("elmer_n_threads", 1)

        if n_processes > 1 and n_threads > 1:
            logging.warning(
                "Using both process and thread level parallelization with Elmer might result in poor performance"
            )

        # for the moment avoid psutil.cpu_count(logical=False)
        max_cpus = int(os.cpu_count() / 2 + 0.5)
        workflow["local_machine_cpu_count"] = max_cpus

        if n_workers == -1:
            n_processes = 1 if n_processes == -1 else n_processes
            n_threads = 1 if n_threads == -1 else n_threads
            n_workers = max(max_cpus // (n_threads * n_processes), 1)
            n_workers = min(n_workers, n_worker_lim)
        elif n_processes == -1:
            n_workers = min(n_workers, n_worker_lim)
            n_threads = 1 if n_threads == -1 else n_threads
            n_processes = max(max_cpus // (n_threads * n_workers), 1)
        elif n_threads == -1:
            n_workers = min(n_workers, n_worker_lim)
            n_threads = max(max_cpus // (n_processes * n_workers), 1)

        requested_cpus = n_workers * n_processes * n_threads
        if requested_cpus > max_cpus:
            logging.warning(f"Requested more CPUs ({requested_cpus}) than available ({max_cpus})")

        gmsh_n_threads = workflow.get("gmsh_n_threads", 1)
        if gmsh_n_threads == -1:
            if parallelization_level == "full_simulation":
                gmsh_n_threads = max(max_cpus // n_workers, 1)
            else:
                gmsh_n_threads = max_cpus

        workflow["n_workers"] = n_workers
        workflow["elmer_n_processes"] = n_processes
        workflow["elmer_n_threads"] = n_threads
        workflow["gmsh_n_threads"] = gmsh_n_threads

    parser = argparse.ArgumentParser()
    parser.add_argument("-q", "--quiet", action="store_true")
    args, _ = parser.parse_known_args()

    if args.quiet:
        workflow.update(
            {
                "run_gmsh_gui": False,  # For GMSH: if true, the mesh is shown after it is done
                # (for large meshes this can take a long time)
                "run_paraview": False,  # this is visual view of the results
            }
        )

    return workflow



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).

import logging
import json
from itertools import product
from pathlib import Path
from shutil import copytree
from typing import Sequence
from scipy.stats import qmc
import numpy as np

from kqcircuits.util.load_save_layout import save_layout
from kqcircuits.util.geometry_json_encoder import GeometryJsonEncoder


def get_combined_parameters(simulation, solution):
    """Return parameters of Simulation and Solution in a combined dictionary.
    In case of common keys, 'solution.' prefix is added to Solution parameter key.
    """
    sim_dict = simulation.get_parameters()
    sol_dict = solution.get_parameters()
    return {
        **{k: v for k, v in sim_dict.items() if k != "name"},
        **{f"solution.{k}" if k in sim_dict else k: v for k, v in sol_dict.items() if k != "name"},
    }


def copy_content_into_directory(source_paths: list, path: Path, folder):
    """Create a folder and copy the contents of the source folders into it

    Arguments:
        source_paths: list of source directories from which to copy content
        path: path where the new folder will be created
        folder: name of the new folder
    """
    if path.exists() and path.is_dir():
        for source_path in source_paths:
            copytree(str(source_path), str(path.joinpath(folder)), dirs_exist_ok=True)


def get_post_process_command_lines(post_process, path, json_filenames):
    """Return post process command line calls as string. Can be used in construction of .bat or .sh script files.

    Args:
        post_process: List of PostProcess objects, a single PostProcess object, or None to be executed after simulations
        path: simulation folder path
        json_filenames: list of paths to simulation json files

    Returns:
        Command lines as string
    """
    if post_process is None:
        return ""

    commands = "echo Post-process\n"
    if isinstance(post_process, list):
        for pp in post_process:
            commands += pp.get_command_line(path, json_filenames)
    else:
        commands += post_process.get_command_line(path, json_filenames)
    return commands


def export_simulation_json(json_data, json_file_path):
    """Export simulation definitions json. Raise an error if file exists"""
    if not Path(json_file_path).exists():
        with open(json_file_path, "w", encoding="utf-8") as fp:
            json.dump(json_data, fp, cls=GeometryJsonEncoder, indent=4)
    else:
        raise ValueError(
            f"Json file '{json_file_path}' already exists. Make sure that simulations and solutions have unique names."
        )


def export_simulation_oas(simulations, path: Path, file_prefix="simulation"):
    """
    Write single OASIS file containing all simulations in list.
    """
    simulations = [simulation[0] if isinstance(simulation, Sequence) else simulation for simulation in simulations]
    unique_layouts = {simulation.layout for simulation in simulations}
    if len(unique_layouts) != 1:
        raise ValueError("Cannot write batch OASIS file since not all simulations are on the same layout.")

    cells = [simulation.cell for simulation in simulations]
    oas_filename = str(path.joinpath(file_prefix + ".oas"))
    save_layout(oas_filename, simulations[0].layout, cells, no_empty_cells=True)
    return oas_filename


def _join_flat_str(value):
    """Returns string in which value is flattened and joined using underscore separator."""
    if isinstance(value, str):
        return value  # return string as it is
    if isinstance(value, dict):
        return _join_flat_str(value.items())  # join keys and values of dictionary
    try:
        return "_".join([_join_flat_str(v) for v in value])  # join terms of any iterable
    except TypeError:
        return str(value)  # convert any non-iterable to string


def sweep_simulation(layout, sim_class, sim_parameters, sweeps):
    """Create simulation sweep by varying one parameter at time. Return list of simulations."""
    simulations = []
    lengths = [len(l) for l in sweeps.values()]
    logging.info(f'Added simulations: {" + ".join([str(l) for l in lengths])} = {sum(lengths)}')
    for param in sweeps:
        for value in sweeps[param]:
            parameters = {
                **sim_parameters,
                param: value,
                "name": _join_flat_str((sim_parameters.get("name", ""), param, value)),
            }
            simulations.append(sim_class(**parameters) if layout is None else sim_class(layout, **parameters))
    return simulations


def cross_sweep_simulation(layout, sim_class, sim_parameters, sweeps):
    """Create simulation sweep by cross-varying all parameters. Return list of simulations."""
    simulations = []
    keys = list(sweeps)
    sets = [list(prod) for prod in product(*sweeps.values())]
    logging.info(f'Added simulations: {" * ".join([str(len(l)) for l in sweeps.values()])} = {len(sets)}')
    for values in sets:
        parameters = {**sim_parameters}
        for i, key in enumerate(keys):
            parameters[key] = values[i]
        parameters["name"] = _join_flat_str((sim_parameters.get("name", ""), values))
        simulations.append(sim_class(**parameters) if layout is None else sim_class(layout, **parameters))
    return simulations


def sweep_solution(sol_class, sol_parameters, sweeps):
    """Create solution sweep by varying one parameter at time. Return list of solutions."""
    return sweep_simulation(None, sol_class, sol_parameters, sweeps)


def cross_sweep_solution(sol_class, sol_parameters, sweeps):
    """Create solution sweep by cross-varying all parameters. Return list of solutions."""
    return cross_sweep_simulation(None, sol_class, sol_parameters, sweeps)


def cross_combine(simulations, solutions):
    """Combines simulations and solutions into list of tuples.

    Args:
        simulations: A Simulation object or a list of Simulation objects.
        solutions: A Solution object or a list of Solution objects.
    Returns:
        A list of tuples containing all combinations of simulations and solutions.

    """
    return list(
        product(
            simulations if isinstance(simulations, Sequence) else [simulations],
            solutions if isinstance(solutions, Sequence) else [solutions],
        )
    )


def unique_rows(a):
    if np.issubdtype(a.dtype, np.number):
        return np.unique(a, axis=0)
    else:
        # Object arrays
        return np.array(list({tuple(row) for row in a}), dtype=object)


def latin_hypercube_sampling(l_bounds, u_bounds, n, integers=True, add_edges=False, remove_duplicates=True):
    """Samples parameters from d-dimensional parameter space defined by bounds using Latin Hypercube sampling.
    Implementation for sampling either integers or floats
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.qmc.LatinHypercube.html

    Args:
        l_bounds: Lower bounds for parameters
        u_bounds: Upper bounds for parameters
        n: Number of samples
        integers: Boolean to check if only integer values should be generated
        add_edges: Boolean to add also the edge points of the hypercube. Adds 2^len(l_bounds) values
        remove_duplicates: Boolen to remove duplicate samples
    Returns:
        A numpy array of samples
    """
    assert len(l_bounds) == len(u_bounds), "Lower and upper bounds have different dimensions"
    sampler = qmc.LatinHypercube(d=len(l_bounds))
    if integers:
        samples = sampler.integers(l_bounds=l_bounds, u_bounds=u_bounds, n=n, endpoint=True)
    else:
        sample = sampler.random(n=n)
        samples = qmc.scale(sample, l_bounds, u_bounds)
    if add_edges:
        # Pair each parameter’s lower and upper bound
        bounds_pairs = list(zip(l_bounds, u_bounds))

        # Generate all combinations of choosing either the lower or upper value
        edge_samples = np.array(list(product(*bounds_pairs)))
        samples = np.vstack((samples, edge_samples))
    if remove_duplicates:
        samples = unique_rows(samples)
    return samples


def latin_hypercube_sampling_mixed_dtypes(
    l_bounds_int, u_bounds_int, l_bounds_float, u_bounds_float, n, add_edges=False
):
    """Samples parameters from d-dimensional parameter space defined by bounds using Latin Hypercube sampling.
    Implemementation for sampling combination of int and float values
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.qmc.LatinHypercube.html

    Args:
        l_bounds_int: Lower bounds for parameters with int dtype
        u_bounds_int: Upper bounds for parameters with int dtype
        l_bounds_float: Lower bounds for parameters with float dtype
        u_bounds_float: Upper bounds for parameters with float dtype
        n: Number of samples
        add_edges: Boolean to add also the edge points of the hypercube. Adds 2^len(l_bounds_int+l_bounds_float) values
    Returns:
        A numpy array of samples
    """
    samples_int = latin_hypercube_sampling(
        l_bounds_int, u_bounds_int, n, integers=True, add_edges=False, remove_duplicates=False
    )
    samples_float = latin_hypercube_sampling(
        l_bounds_float, u_bounds_float, n, integers=False, add_edges=False, remove_duplicates=False
    )
    samples = np.hstack((samples_int, samples_float), dtype=object)
    if add_edges:
        # Pair each parameter’s lower and upper bound
        bounds_pairs = list(zip(l_bounds_int + l_bounds_float, u_bounds_int + u_bounds_float))

        # Generate all combinations of choosing either the lower or upper value
        edge_samples = np.array(list(product(*bounds_pairs)), dtype=object)
        samples = np.vstack((samples, edge_samples))
    samples = unique_rows(samples)
    return samples


def combine_sweep_simulation(layout, sim_class, sim_parameters, keys, samples):
    """Creates a simulation sweep from created samples. Returns a list of simulations.
    Args:
        layout: Layout for simulation
        sim_class: Simulation class
        sim_parameters: Simulation parameters which are not in the sweep
        keys: List of parameter names corresponding to samples
        samples: List of samples for sweep parameters
    Returns:
        A list of simulations
    """
    assert len(samples[0]) == len(keys), "Samples don't have same dimension as keys"
    simulations = []
    for values in samples:
        parameters = {**sim_parameters}
        for i, key in enumerate(keys):
            parameters[key] = values[i]
        parameters["name"] = _join_flat_str((sim_parameters.get("name", ""), values))
        simulations.append(sim_class(**parameters) if layout is None else sim_class(layout, **parameters))
    return simulations



# This code is part of KQCircuits
# Copyright (C) 2025 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).
from kqcircuits.pya_resolver import pya


def load_layout(filename, layout: pya.Layout, **opts) -> None:
    """Loads ``Layout`` from file.

    The default LoadLayoutOptions of KLayout are employed with following exceptions:
    * This function sets cell_conflict_resolution = RenameCell by default (conflicting cells will be renamed).
    * The LoadLayoutOptions can be modified via keyword arguments.

    Args:
        filename: The name of the file to load.
        layout: The layout object into which the file is loaded.
        opts: Custom LoadLayoutOptions as keyword arguments.
    """
    load_opts = pya.LoadLayoutOptions()
    load_opts.cell_conflict_resolution = pya.LoadLayoutOptions.CellConflictResolution.RenameCell
    for key, value in opts.items():
        if not hasattr(load_opts, key):
            raise NotImplementedError(f"pya.LoadLayoutOptions has no attribute called {key}.")
        setattr(load_opts, key, value)
    layout.read(str(filename), load_opts)


def save_layout(
    filename,
    layout: pya.Layout,
    cells: list[pya.Cell] | None = None,
    layers: list[pya.LayerInfo] | None = None,
    **opts,
) -> None:
    """Saves ``Layout`` to file.

    The default SaveLayoutOptions of KLayout are employed with following exceptions:
    * This function calls set_format_from_filename to select file format according to the file's extension.
    * By default write_context_info = False to enable saving cells as static cells.
    * If cells or layers are specified, only given instances will be saved.
    * The SaveLayoutOptions can be modified via keyword arguments.

    Args:
        filename: The name of the file to save.
        layout: The layout object whose geometry will be saved.
        cells: List of pya.Cell objects to indicate which cells are saved. Save all cells if not specified.
        layers: List of pya.LayerInfo objects to indicate which layers are saved. Save all layers if not specified.
        opts: Custom SaveLayoutOptions as keyword arguments.
    """
    save_opts = pya.SaveLayoutOptions()
    save_opts.set_format_from_filename(filename)
    save_opts.write_context_info = False
    if cells is not None:
        save_opts.clear_cells()
        for cell in cells:
            save_opts.add_cell(cell.cell_index())
    if layers is not None:
        save_opts.deselect_all_layers()
        for layer in layers:
            save_opts.add_layer(layout.layer(layer), layer)
    for key, value in opts.items():
        if not hasattr(save_opts, key):
            raise NotImplementedError(f"pya.SaveLayoutOptions has no attribute called {key}.")
        setattr(save_opts, key, value)
    layout.write(str(filename), save_opts)



# This code is part of KQCircuits
# Copyright (C) 2025 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).

from kqcircuits.pya_resolver import pya
from kqcircuits.simulations.epr.util import EPRTarget, create_bulk_and_mer_partition_regions, is_flip_chip
from kqcircuits.simulations.partition_region import PartitionRegion

# Partition region and correction cuts definitions for double_pads qubit

metal_edge_dimension = 3.0
vertical_dimension = 3.0
metal_edge_dimensions = [
    metal_edge_dimension,  # gap
    metal_edge_dimension,  # metal
]


def partition_regions(simulation: EPRTarget, prefix: str = "") -> list[PartitionRegion]:
    base = simulation.refpoints["base"]

    island1_extent_0 = float(simulation.island1_extent[0])
    island1_extent_1 = float(simulation.island1_extent[1])
    island2_extent_0 = float(simulation.island2_extent[0])
    island2_extent_1 = float(simulation.island2_extent[1])

    junction1 = simulation.refpoints["probe_island_1"] - pya.DPoint(0, island1_extent_1 / 2)
    junction2 = simulation.refpoints["probe_island_2"] + pya.DPoint(0, island2_extent_1 / 2)

    squid_height = junction1.y - base.y
    island1_bottom = simulation.squid_offset + squid_height / 2
    taper_height = (simulation.island_island_gap - squid_height) / 2

    island1_polygon = pya.DPolygon(
        [
            pya.DPoint(
                -island1_extent_0 / 2 - metal_edge_dimension,
                island1_bottom + taper_height + island1_extent_1 + metal_edge_dimension,
            ),
            pya.DPoint(
                island1_extent_0 / 2 + metal_edge_dimension,
                island1_bottom + taper_height + island1_extent_1 + metal_edge_dimension,
            ),
            pya.DPoint(
                island1_extent_0 / 2 + metal_edge_dimension, island1_bottom + taper_height - metal_edge_dimension
            ),
            pya.DPoint(
                -island1_extent_0 / 2 - metal_edge_dimension, island1_bottom + taper_height - metal_edge_dimension
            ),
        ]
    )

    island1_region = pya.Region(
        island1_polygon.transformed(pya.DCplxTrans(1, 0, False, base.x, base.y)).to_itype(simulation.layout.dbu)
    )

    island1_region.round_corners(
        simulation.island1_r / simulation.layout.dbu, simulation.island1_r / simulation.layout.dbu, simulation.n
    )

    island2_top = simulation.squid_offset - squid_height / 2
    island2_polygon = pya.DPolygon(
        [
            pya.DPoint(
                -island2_extent_0 / 2 - metal_edge_dimension,
                island2_top - taper_height - island2_extent_1 - metal_edge_dimension,
            ),
            pya.DPoint(
                island2_extent_0 / 2 + metal_edge_dimension,
                island2_top - taper_height - island2_extent_1 - metal_edge_dimension,
            ),
            pya.DPoint(island2_extent_0 / 2 + metal_edge_dimension, island2_top - taper_height + metal_edge_dimension),
            pya.DPoint(-island2_extent_0 / 2 - metal_edge_dimension, island2_top - taper_height + metal_edge_dimension),
        ]
    )

    island2_region = pya.Region(
        island2_polygon.transformed(pya.DCplxTrans(1, 0, False, base.x, base.y)).to_itype(simulation.layout.dbu)
    )
    island2_region.round_corners(
        simulation.island2_r / simulation.layout.dbu, simulation.island2_r / simulation.layout.dbu, simulation.n
    )

    sector_region = island1_region + island2_region

    # Lead region
    half_leads_w = max(simulation.island1_taper_width, simulation.island2_taper_width) / 2 + metal_edge_dimension
    leads_p1 = junction1 - pya.DPoint(half_leads_w, 0)
    leads_p2 = junction2 + pya.DPoint(half_leads_w, 0)

    # coupler region
    first_island_top_edge = (simulation.squid_offset + squid_height / 2) + taper_height + island1_extent_1
    coupler_top_edge = first_island_top_edge + simulation.coupler_offset + float(simulation.coupler_extent[1])

    coupler_path_polygon = pya.DPolygon(
        [
            pya.DPoint(
                -simulation.coupler_a / 2 - 2 * metal_edge_dimension, (float(simulation.ground_gap[1]) / 2 + 100)
            ),
            pya.DPoint(
                simulation.coupler_a / 2 + 2 * metal_edge_dimension, (float(simulation.ground_gap[1]) / 2 + 100)
            ),
            pya.DPoint(
                simulation.coupler_a / 2 + 2 * metal_edge_dimension, coupler_top_edge + 2 * metal_edge_dimension
            ),
            pya.DPoint(
                -simulation.coupler_a / 2 - 2 * metal_edge_dimension, coupler_top_edge + 2 * metal_edge_dimension
            ),
        ]
    )

    coupler1_region = pya.Region(
        coupler_path_polygon.transformed(pya.DCplxTrans(1, 0, False, base.x, base.y)).to_itype(simulation.layout.dbu)
    )

    coupler2_polygon = pya.DPolygon(
        [
            pya.DPoint(
                -float(simulation.coupler_extent[0]) / 2 - 2 * metal_edge_dimension,
                coupler_top_edge + 2 * metal_edge_dimension,
            ),
            pya.DPoint(
                -float(simulation.coupler_extent[0]) / 2 - 2 * metal_edge_dimension,
                first_island_top_edge + simulation.coupler_offset - 2 * metal_edge_dimension,
            ),
            pya.DPoint(
                float(simulation.coupler_extent[0]) / 2 + 2 * metal_edge_dimension,
                first_island_top_edge + simulation.coupler_offset - 2 * metal_edge_dimension,
            ),
            pya.DPoint(
                float(simulation.coupler_extent[0]) / 2 + 2 * metal_edge_dimension,
                coupler_top_edge + 2 * metal_edge_dimension,
            ),
        ]
    )

    coupler2_region = pya.Region(
        coupler2_polygon.transformed(pya.DCplxTrans(1, 0, False, base.x, base.y)).to_itype(simulation.layout.dbu)
    )
    coupler2_region.round_corners(
        simulation.coupler_r / simulation.layout.dbu, simulation.coupler_r / simulation.layout.dbu, simulation.n
    )

    result = create_bulk_and_mer_partition_regions(
        name=f"{prefix}coupler1",
        face=simulation.face_ids[0],
        region=coupler1_region,
        vertical_dimensions=vertical_dimension,
        metal_edge_dimensions=metal_edge_dimension,
        bulk=False,
        visualise=True,
    )
    result += create_bulk_and_mer_partition_regions(
        name=f"{prefix}coupler2",
        face=simulation.face_ids[0],
        region=coupler2_region,
        vertical_dimensions=vertical_dimension,
        metal_edge_dimensions=metal_edge_dimension,
        bulk=False,
        visualise=True,
    )
    result += create_bulk_and_mer_partition_regions(
        name=f"{prefix}island",
        face=simulation.face_ids[0],
        region=sector_region,
        vertical_dimensions=vertical_dimension,
        # Hack for preventing a thin region of islandbulk from forming due to rounding (discretization).
        # Makes the MER larger on the gap side to cover this region.
        metal_edge_dimensions=[6.0, metal_edge_dimensions[1]],
        visualise=True,
    )
    result += create_bulk_and_mer_partition_regions(
        name=f"{prefix}leads",
        face=simulation.face_ids[0],
        metal_edge_dimensions=metal_edge_dimensions,
        region=pya.DBox(leads_p1, leads_p2),
        vertical_dimensions=vertical_dimension,
        visualise=True,
    )
    result += create_bulk_and_mer_partition_regions(
        name=f"{prefix}{simulation.face_ids[0]}complement",
        face=simulation.face_ids[0],
        metal_edge_dimensions=metal_edge_dimensions,
        region=None,
        vertical_dimensions=vertical_dimension,
        visualise=True,
    )
    if is_flip_chip(simulation) and simulation.etch_opposite_face:
        result += create_bulk_and_mer_partition_regions(
            name=f"{prefix}{simulation.face_ids[1]}complement",
            face=simulation.face_ids[1],
            metal_edge_dimensions=metal_edge_dimensions,
            region=None,
            vertical_dimensions=vertical_dimension,
            visualise=True,
        )

    return result


def correction_cuts(simulation: EPRTarget, prefix: str = "") -> dict[str, dict]:
    half_cut_length = 30.0
    base = simulation.refpoints["base"]
    probe1 = simulation.refpoints["probe_island_1"]

    leads_center = simulation.refpoints.get("squid_port_squid_a", base) + pya.DPoint(0, 1.5 * metal_edge_dimension)

    island1_extent_0 = float(simulation.island1_extent[0])
    island1_extent_1 = float(simulation.island1_extent[1])
    coupler_extent0 = float(simulation.coupler_extent[0])
    coupler_extent1 = float(simulation.coupler_extent[1])
    ground_gap = base + pya.DPoint(float(simulation.ground_gap[0]) / 2, 0)
    island1_edge = base + pya.DPoint(island1_extent_0 / 4.0, simulation.island_island_gap / 2.0)
    coupler2_center = probe1 + pya.DPoint(0, island1_extent_1 / 2 + simulation.coupler_offset + coupler_extent1 / 2)
    coupler1_center = probe1 + pya.DPoint(0, island1_extent_1 / 2 + simulation.coupler_offset + 4 * coupler_extent1)

    result = {
        f"{prefix}islandmer": {
            "p1": island1_edge + pya.DPoint(0, -half_cut_length),
            "p2": island1_edge + pya.DPoint(0, half_cut_length),
            "boundary_conditions": {"xmin": {"potential": 0}},
        },
        f"{prefix}leadsmer": {
            "p1": leads_center + pya.DPoint(-half_cut_length, 0),
            "p2": leads_center + pya.DPoint(half_cut_length, 0),
            "boundary_conditions": {"xmax": {"potential": 0}, "xmin": {"potential": 0}},
        },
        f"{prefix}coupler2mer": {
            "p1": coupler2_center + pya.DPoint(coupler_extent0 / 4.0, -half_cut_length),
            "p2": coupler2_center + pya.DPoint(coupler_extent0 / 4.0, half_cut_length),
            "boundary_conditions": {"xmax": {"potential": 0}, "xmin": {"potential": 0}},
        },
        f"{prefix}coupler1mer": {
            "p1": coupler1_center + pya.DPoint(-half_cut_length, 0),
            "p2": coupler1_center + pya.DPoint(half_cut_length, 0),
            "boundary_conditions": {"xmax": {"potential": 0}, "xmin": {"potential": 0}},
        },
        f"{prefix}{simulation.face_ids[0]}complementmer": {
            "p1": ground_gap + pya.DPoint(-half_cut_length, 0),
            "p2": ground_gap + pya.DPoint(half_cut_length, 0),
            "boundary_conditions": {"xmin": {"potential": 1}},
        },
    }
    if is_flip_chip(simulation) and simulation.etch_opposite_face:
        result[f"{prefix}{simulation.face_ids[1]}complementmer"] = {
            "p1": ground_gap + pya.DPoint(-half_cut_length, 0),
            "p2": ground_gap + pya.DPoint(half_cut_length, 0),
            "boundary_conditions": {"xmin": {"potential": 1}},
        }

    return result



# This code is part of KQCircuits
# Copyright (C) 2024 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).

from kqcircuits.pya_resolver import pya
from kqcircuits.elements.smooth_capacitor import SmoothCapacitor
from kqcircuits.elements.finger_capacitor_square import eval_a2, eval_b2
from kqcircuits.simulations.epr.util import in_gui, EPRTarget, get_mer_z, create_bulk_and_mer_partition_regions
from kqcircuits.simulations.partition_region import PartitionRegion


# Partition region and correction cuts definitions for Swissmon qubit
vertical_dimension = 1.0
metal_edge_dimension = 2.0


def partition_regions(simulation: EPRTarget, prefix: str = "") -> list[PartitionRegion]:

    port_a_rf = simulation.refpoints["port_a"]
    port_b_rf = simulation.refpoints["port_b"]

    a2, b2 = eval_a2(simulation), eval_b2(simulation)

    # Make a stub SmoothCapacitor instance, copy some attributes
    # then use some region generating functions to create a partition region shape overlapping finger edges
    s = SmoothCapacitor()
    s.n = simulation.n
    s.finger_width = simulation.finger_width
    s.finger_gap = simulation.finger_gap
    s.finger_control = simulation.finger_control
    s.ground_gap = simulation.ground_gap
    s.layout = simulation.layout

    right_fingers, left_fingers = s.get_finger_regions()
    region_ground = right_fingers + left_fingers
    region_ground += s.middle_gap_fill().sized(-simulation.ground_gap / simulation.layout.dbu, 5)
    fingers = (
        s.super_smoothen_region(region_ground, simulation.finger_gap + simulation.ground_gap)
        .sized(-(simulation.finger_width / 2.0) / simulation.layout.dbu, 5)
        .transformed(pya.DTrans(0, False, simulation.refpoints["base"]).to_itype(simulation.layout.dbu))
    )

    result = []
    if not in_gui(simulation):
        port_a_len = 11 + simulation.waveguide_length + metal_edge_dimension  # 11 is the hardcoded port dimension
        port_a_width = simulation.a + 2 * simulation.b + 2 * metal_edge_dimension
        port_a_middle = port_a_rf - pya.DPoint(port_a_len / 2.0, 0)
        port_a_dp = pya.DPoint(port_a_len / 2.0, port_a_width / 2)

        port_b_len = 11 + simulation.waveguide_length + metal_edge_dimension  # 11 is the hardcoded port dimension
        port_b_width = a2 + 2 * b2 + 2 * metal_edge_dimension
        port_b_middle = port_b_rf + pya.DPoint(port_b_len / 2.0, 0)
        port_b_dp = pya.DPoint(port_b_len / 2.0, port_b_width / 2)

        if simulation.use_internal_ports:
            port_a_region = pya.DBox(port_a_middle - port_a_dp, port_a_middle + port_a_dp)
            port_b_region = pya.DBox(port_b_middle - port_b_dp, port_b_middle + port_b_dp)
        else:
            port_a_dpx_edge = pya.DPoint(-simulation.box.width() / 2.0, 0.0)
            port_a_dpy = pya.DPoint(0.0, port_a_width / 2)
            port_a_middley = pya.DPoint(0, port_a_middle.y)
            port_a_region = pya.DBox(port_a_dpx_edge + port_a_middley - port_a_dpy, port_a_middle + port_a_dp)

            port_b_dpx_edge = pya.DPoint(simulation.box.width(), 0.0)
            port_b_dpy = pya.DPoint(0.0, port_b_width / 2)
            port_b_middley = pya.DPoint(0, port_b_middle.y)
            port_b_region = pya.DBox(port_b_middle - port_b_dp, port_b_dpx_edge + port_b_middley + port_b_dpy)

        result += create_bulk_and_mer_partition_regions(
            name=f"{prefix}port_b",
            face=simulation.face_ids[0],
            metal_edge_dimensions=metal_edge_dimension,
            region=port_b_region,
            vertical_dimensions=vertical_dimension,
            visualise=True,
        )

        result += create_bulk_and_mer_partition_regions(
            name=f"{prefix}port_a",
            face=simulation.face_ids[0],
            metal_edge_dimensions=metal_edge_dimension,
            region=port_a_region,
            vertical_dimensions=vertical_dimension,
            visualise=True,
        )
    result += create_bulk_and_mer_partition_regions(
        name=f"{prefix}fingers",
        face=simulation.face_ids[0],
        metal_edge_dimensions=metal_edge_dimension,
        region=fingers,
        vertical_dimensions=vertical_dimension,
        visualise=True,
    )
    result += create_bulk_and_mer_partition_regions(
        name=f"{prefix}{simulation.face_ids[0]}complement",
        face=simulation.face_ids[0],
        metal_edge_dimensions=metal_edge_dimension,
        region=None,
        vertical_dimensions=vertical_dimension,
        visualise=True,
    )
    if in_gui(simulation):
        make_t_regions = True
    else:
        make_t_regions = len(simulation.face_stack) > 1
    if make_t_regions:
        result += create_bulk_and_mer_partition_regions(
            name=f"{prefix}{simulation.face_ids[1]}complement",
            face=simulation.face_ids[1],
            metal_edge_dimensions=metal_edge_dimension,
            region=None,
            vertical_dimensions=vertical_dimension,
            visualise=True,
        )

    return result


def correction_cuts(simulation: EPRTarget, prefix: str = "") -> dict[str, dict]:
    base_rf = simulation.refpoints["base"]
    port_a_rf = simulation.refpoints["port_a"]
    port_b_rf = simulation.refpoints["port_b"]

    a2, b2 = eval_a2(simulation), eval_b2(simulation)

    gaps = pya.Region(simulation.cell.begin_shapes_rec(simulation.get_layer("base_metal_gap_wo_grid")))
    finger_top = None
    finger_center = None
    for polygon in gaps.each():
        for edge in polygon.to_dtype(simulation.layout.dbu).each_edge():
            # Hole edges are oriented counterclockwise while hull edges are oriented clockwise.
            # Rely on this orientation to get top point of fingers.
            if edge.d().x < 0.0 and (finger_top is None or edge.p1.y > finger_top.y):
                finger_top = edge.p1

            k = edge.d().sprod(base_rf - edge.p1) / edge.d().sq_abs()
            nearest = (edge.p1 if k <= 0.0 else (edge.p2 if k >= 1.0 else edge.p1 + k * edge.d())) - base_rf
            if finger_center is None or nearest.abs() < finger_center.abs():
                if simulation.finger_control > 1.0 or abs(nearest.y) < abs(nearest.x):
                    finger_center = nearest

    if not in_gui(simulation):
        port_a_len = 11 + simulation.waveguide_length + metal_edge_dimension  # 11 is the hardcoded port dimension
        port_a_middle = port_a_rf - pya.DPoint(port_a_len / 2.0, 0)
        port_a_width = simulation.a + 2 * simulation.b + 2 * metal_edge_dimension

        port_b_len = 11 + simulation.waveguide_length + metal_edge_dimension  # 11 is the hardcoded port dimension
        port_b_middle = port_b_rf + pya.DPoint(port_b_len / 2.0, 0)
        port_b_width = a2 + 2 * b2 + 2 * metal_edge_dimension

    z_me = get_mer_z(simulation, simulation.face_ids[0])
    fingersmer_end = finger_center * (1.0 + simulation.finger_width * 0.9 / finger_center.abs())
    half_cut_len = 25.0
    result = {
        f"{prefix}fingersmer": {
            "p1": base_rf + fingersmer_end,
            "p2": base_rf - fingersmer_end,
            "metal_edges": [
                {"x": simulation.finger_width * 0.9, "z": z_me},
                {"x": 2 * fingersmer_end.abs() - simulation.finger_width * 0.9, "z": z_me},
            ],
        },
        f"{prefix}{simulation.face_ids[0]}complementmer": {
            "p1": finger_top + pya.DVector(0, simulation.ground_gap + half_cut_len),
            "p2": finger_top - pya.DVector(0, simulation.finger_width * 0.9),
            "metal_edges": [
                {"x": half_cut_len, "z": z_me},
                {"x": half_cut_len + simulation.ground_gap, "z": z_me},
            ],
        },
    }
    if not in_gui(simulation):
        result[f"{prefix}port_amer"] = {
            "p1": port_a_middle - pya.DPoint(0, port_a_width),
            "p2": port_a_middle + pya.DPoint(0, port_a_width),
            "metal_edges": [
                {
                    "x": port_a_width - simulation.a / 2.0 - simulation.b,
                    "z": z_me,
                },
                {
                    "x": port_a_width - simulation.a / 2.0,
                    "z": z_me,
                },
                {
                    "x": port_a_width + simulation.a / 2.0,
                    "z": z_me,
                },
                {
                    "x": port_a_width + simulation.a / 2.0 + simulation.b,
                    "z": z_me,
                },
            ],
        }
        result[f"{prefix}port_bmer"] = {
            "p1": port_b_middle - pya.DPoint(0, port_b_width),
            "p2": port_b_middle + pya.DPoint(0, port_b_width),
            "metal_edges": [
                {"x": port_b_width - a2 / 2 - b2, "z": z_me},
                {"x": port_b_width - a2 / 2, "z": z_me},
                {"x": port_b_width + a2 / 2, "z": z_me},
                {"x": port_b_width + a2 / 2 + b2, "z": z_me},
            ],
        }
    return result



# This code is part of KQCircuits
# Copyright (C) 2024 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).


from math import pi, cos, sqrt

from kqcircuits.elements.finger_capacitor_square import FingerCapacitorSquare, eval_a2, eval_b2
from kqcircuits.elements.smooth_capacitor import SmoothCapacitor, unit_vector, segment_points
from kqcircuits.pya_resolver import pya
from kqcircuits.util.geometry_helper import get_angle
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.elements.element import Element


@add_parameters_from(SmoothCapacitor, "a2", "b2", "finger_width", "finger_gap", "ground_gap")
class SpiralCapacitor(Element):
    """The PCell declaration for a spiral capacitor."""

    finger_width2 = Param(
        pdt.TypeDouble,
        "Width of a finger on the right",
        -1,
        unit="μm",
        docstring="Non-physical value '-1' means that same width is used on both sides.",
    )
    ground_gap2 = Param(
        pdt.TypeDouble,
        "Gap between ground and finger on the right",
        -1,
        unit="μm",
        docstring="Non-physical value '-1' means that same gap is used on both sides.",
    )

    spiral_angle = Param(pdt.TypeDouble, "Angle of spiral in degrees", 180, unit="degrees")
    spiral_angle2 = Param(
        pdt.TypeDouble,
        "Angle of spiral in degrees on the right",
        -1,
        unit="degrees",
        docstring="Non-physical value '-1' means that same angle is used on both sides.",
    )

    rotation = Param(
        pdt.TypeDouble,
        "Waveguide rotation in degrees",
        -1,
        unit="degrees",
        docstring="Non-physical value '-1' means that the waveguide points towards the spiral center.",
    )
    rotation2 = Param(
        pdt.TypeDouble,
        "Waveguide rotation in degrees on the right",
        -1,
        unit="degrees",
        docstring="Non-physical value '-1' means that the same rotation is used on both sides.",
    )

    def can_create_from_shape_impl(self):
        return self.shape.is_path()

    def build(self):

        def spiral_segment(segment_index, spiral_angle, rotation):
            finger_width2 = self.finger_width if self.finger_width2 < 0 else self.finger_width2
            width = (self.finger_width + finger_width2) / 2 + self.finger_gap
            angle = (segment_index % 2 + rotation / 180) * pi
            origin = pya.DPoint(-unit_vector(angle) * width / 2)
            turn = (spiral_angle / 180 - segment_index) * pi
            radius = (segment_index + 1) * width
            return origin, angle, turn, radius

        def spiral_position(spiral_angle):
            origin, angle, turn, radius = spiral_segment(int(spiral_angle / 180), spiral_angle, 0.0)
            position = origin + unit_vector(angle + turn) * radius
            return get_angle(position), position.abs()

        def spiral_region(spiral_angle, rotation, finger_r):
            pnts = []
            stnp = []
            segments = int(spiral_angle / 180 - 1e-3)
            for s in range(segments + 1):
                o, a, t, r = spiral_segment(s, spiral_angle if s == segments else 180 * (s + 1), rotation)
                v0 = unit_vector(a)
                v1 = unit_vector(a + t)
                if s == 0:
                    pnts += segment_points(o + v0 * (r + finger_r), a - pi / 2, -pi, pi * finger_r, self.n)
                pnts += segment_points(o + v0 * (r - finger_r), a + pi / 2, t, t * (r - finger_r), self.n)
                stnp += segment_points(o + v1 * (r + finger_r), a + t - pi / 2, -t, t * (r + finger_r), self.n)[::-1]
                if s == segments:
                    pnts += segment_points(o + v1 * (r - finger_r), a + t + pi / 2, -pi, pi * finger_r, self.n)
            return pya.Region(pya.DPolygon(pnts + stnp[::-1]).to_itype(self.layout.dbu))

        def wg_region(p0, r0, p1, w1, l1=0):
            l0 = (p1 - p0).abs()
            dx = (p1 - p0) / l0
            dy = pya.DVector(dx.y, -dx.x)
            div = r0 * r0 * l0 / (w1 * w1 + l0 * l0)
            rx = (1 - sqrt(1 - (r0 * r0 - w1 * w1) / (l0 * div))) * div
            ry = sqrt(r0 * r0 - rx * rx)
            pnts = [p1 - dy * w1, p0 + dx * rx - dy * ry, p0 + dx * rx + dy * ry, p1 + dy * w1]
            if l1 > 0:
                pnts += [p1 + dx * l1 + dy * w1, p1 + dx * l1 - dy * w1]
            return pya.Region(pya.DPolygon(pnts).to_itype(self.layout.dbu))

        def port_position(position, direction, distance):
            return position + (distance - direction.sprod(position.to_v())) * direction

        # Process terms on the right
        a2, b2 = eval_a2(self), eval_b2(self)
        finger_width2 = self.finger_width if self.finger_width2 < 0 else self.finger_width2
        ground_gap2 = self.ground_gap if self.ground_gap2 < 0 else self.ground_gap2
        spiral_angle2 = self.spiral_angle if self.spiral_angle2 < 0 else self.spiral_angle2
        rotation2 = self.rotation if self.rotation2 < 0 else self.rotation2

        # Find angles for spirals
        rot, length = spiral_position(self.spiral_angle - max(0, self.rotation - 90))
        spiral_rotation = 180 - rot if self.rotation < 0 else 90 - self.spiral_angle + self.rotation
        rot2, length2 = spiral_position(spiral_angle2 - max(0, rotation2 - 90))
        spiral_rotation2 = spiral_rotation - 180
        rotation_out = spiral_rotation2 + (rot2 if rotation2 < 0 else 90 + spiral_angle2 - rotation2)

        # Create spiral regions
        finger_left = spiral_region(self.spiral_angle, spiral_rotation, self.finger_width / 2)
        finger_right = spiral_region(spiral_angle2, spiral_rotation2, finger_width2 / 2)

        # Find waveguide positions and directions
        pos = pya.DPoint(length * unit_vector((rot + spiral_rotation) / 180 * pi))
        port_dir = pya.DVector(-1, 0)
        port_pos = port_position(pos, port_dir, length + self.finger_width / 2 + self.ground_gap)
        pos2 = pya.DPoint(length2 * unit_vector((rot2 + spiral_rotation2) / 180 * pi))
        port_dir2 = unit_vector(rotation_out / 180 * pi)
        port_pos2 = port_position(pos2, port_dir2, length2 + finger_width2 / 2 + ground_gap2)

        # Create waveguide regions
        safe_scale = cos(pi / self.n)
        wg_left = wg_region(pos, self.finger_width / 2 * safe_scale, port_pos, self.a / 2, length)
        gap_left = wg_region(pos, (self.finger_width / 2 + self.ground_gap) * safe_scale, port_pos, self.a / 2 + self.b)
        wg_right = wg_region(pos2, finger_width2 / 2 * safe_scale, port_pos2, a2 / 2, length2)
        gap_right = wg_region(pos2, (finger_width2 / 2 + ground_gap2) * safe_scale, port_pos2, a2 / 2 + b2)

        # Combine gap and metal regions
        region_ground = (
            finger_left.sized(self.ground_gap / self.layout.dbu, 5)
            + finger_right.sized(ground_gap2 / self.layout.dbu, 5)
            + gap_left
            + gap_right
        )
        gap_region = region_ground - finger_left - finger_right - wg_left - wg_right

        # Create shapes into cell
        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(gap_region)

        # protection
        region_protection = region_ground.sized(self.margin / self.layout.dbu, 5)
        self.add_protection(region_protection)

        # Add size into annotation layer
        self.cell.shapes(self.get_layer("annotations")).insert(pya.DText(str(round(self.spiral_angle, 5)), 0, 0))

        # Create ports
        self.add_port("a", port_pos, port_dir)
        self.add_port("b", port_pos2, port_dir2)

    @classmethod
    def get_sim_ports(cls, simulation):
        return FingerCapacitorSquare.get_sim_ports(simulation)



# This code is part of KQCircuits
# Copyright (C) 2025 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).
from typing import Callable
from itertools import zip_longest

from kqcircuits.pya_resolver import pya
from kqcircuits.util.node import Node


def coerce_nodes_with_gui_path(
    nodes: list[Node],
    gui_path: pya.DPath,
    gui_path_shadow: pya.DPath,
    snap_point_callback: Callable[[pya.DPoint], pya.DPoint],
    dbu,
) -> tuple[list[Node], pya.DPath]:
    """Detect GUI editing changes in WaveguideComposite-like elements, and update the nodes list accordingly.

    To implement GUI editing of a WaveguiceComposite-like element, it needs three parameters:
    - ``nodes``: the actual list of ``Node`` objects
    - ``gui_path``: A ``pya.DPath`` parameter which is editable by the user in KLayout
    - ``gui_path_shadow``: A ``pya.DPath`` parameter which is hidden, and thus not edited by the user

    This function detects changes between ``gui_path`` and ``gui_path_shadow``. If there are changes, ``nodes`` is
    updated to reflect the coordinates.

    To use this function, call it in ``coerce_parameters_impl`` and use the output to update all three element
    parameters.

    Args:
        nodes: The ``Node`` objects (deserialzed with Node.nodes_from_string)
        gui_path: Path with a point for each Node that may have been edited by the user
        gui_path_shadow: reference Path with a point for each user that is never edited by the user.
        snap_point_callback: callback that can modify points to snap to a relevant grid. Only applied to new and
            changed positions.
        dbu: Value of layout.dbu

    Returns: tuple containing:

    - ``new_nodes``, updated list of Nodes, update the ``nodes`` parameter to this value (converted to strings)
    - ``new_path``, updated path, update ``gui_path`` and ``gui_path_shadow`` to this value
    """
    if len(nodes) < 2 or gui_path == gui_path_shadow:
        # Force rounding to integer database units since the KLayout Partial tool also rounds to database units.
        new_path = gui_path.to_itype(dbu).to_dtype(dbu)
        return nodes, new_path

    new_points = list(gui_path.each_point())
    old_points = list(gui_path_shadow.each_point())

    length_change = len(new_points) - len(old_points)
    if length_change == 0:
        # One or more points were moved; update node positions of the moved points
        for node, new_position, old_position in zip(nodes, new_points, old_points):
            if new_position != old_position:
                new_position = snap_point_callback(new_position)
                node.position = new_position
    elif length_change == 1:
        # One node was added. Figure out which node it was.
        added_index = None
        added_point = None
        for i, (new_point, old_point) in enumerate(zip_longest(new_points, old_points)):
            if old_point is None or new_point != old_point:
                added_index = i
                added_point = new_point
                break

        if added_index is not None:
            nodes.insert(added_index, Node(snap_point_callback(added_point)))
    elif length_change < 0 and gui_path.num_points() >= 2:
        # One or more points deleted; delete corresponding nodes. We require at least two remaining nodes.
        remaining_indices = []
        new_indices = []

        for i, old_point in enumerate(old_points):
            if old_point in new_points:
                remaining_indices.append(i)
                new_indices.append(new_points.index(old_point))

        # Verify we found the right number of remaining indices, and the order is correct
        new_indices_monotonic = all(i < j for i, j in zip(new_indices, new_indices[1:]))
        if len(remaining_indices) == len(new_points) and new_indices_monotonic:
            nodes = [nodes[i] for i in remaining_indices]

    # After coerce the gui_path and gui_path_shadow should both match the nodes.
    new_path = pya.DPath([node.position for node in nodes], 1)
    # Force rounding to integer database units since the KLayout Partial tool also rounds to database units.
    new_path = new_path.to_itype(dbu).to_dtype(dbu)
    return nodes, new_path



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).


from math import pi, cos, sin, atan
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.elements.element import Element
from kqcircuits.elements.finger_capacitor_square import FingerCapacitorSquare, eval_a2, eval_b2


def unit_vector(radians):
    return pya.DVector(cos(radians), sin(radians))


def segment_points(start_pos, start_ang, turn, length, n):
    if length == 0:
        return []
    if turn == 0:
        return [start_pos + length * unit_vector(start_ang)]
    r = length / turn  # signed radius
    center = start_pos + r * unit_vector(start_ang + pi / 2)  # center of the turn circle
    num_pnts = max(round(abs(turn) * n / (2 * pi)), 1)  # number of new points
    return [center - r * unit_vector(start_ang + pi / 2 + turn * (i + 1) / num_pnts) for i in range(num_pnts)]


@add_parameters_from(FingerCapacitorSquare, "fixed_length", "a2", "b2", finger_width=10, finger_gap=5)
class SmoothCapacitor(Element):
    """The PCell declaration for a smooth finger capacitor.

    SmoothCapacitor is a finger capacitor, which has continuous geometry changes
    through the capacitance range. This leads to continuous capacitance function,
    which enables using capacitor inside numerical optimization methods.

    Capacitance range is achieved by changing single parameter called `finger_control`.
    """

    finger_control = Param(
        pdt.TypeDouble,
        "Continuously adjust finger number",
        2.1,
        docstring="Parameter for capacitor growth (related to number of fingers per side)",
    )
    ground_gap = Param(pdt.TypeDouble, "Gap between ground and finger", 10, unit="μm")

    def can_create_from_shape_impl(self):
        return self.shape.is_path()

    def t_poly(self, bend, length):
        r1 = self.finger_width / 2
        r2 = r1 + self.finger_gap
        # Bottom 180-degree bend starting from origin
        ang = 3 * pi / 2
        pnts = segment_points(pya.DPoint(0, 0), ang, -pi, pi * r1, self.n)
        ang -= pi
        # Bend before straight segment
        bend0 = min(bend, pi / 2)
        pnts += segment_points(pnts[-1], ang, bend0, bend0 * r2, self.n)
        ang += bend0
        # Straight segment, 180-degree bend, and straight segment back
        pnts += segment_points(pnts[-1], ang, 0.0, length, self.n)
        pnts += segment_points(pnts[-1], ang, -pi, pi * r1, self.n)
        ang -= pi
        pnts += segment_points(pnts[-1], ang, 0.0, length, self.n)
        # Possible turn upwards if bend > pi/2
        turn = bend - pi / 2
        if turn > 0.0:
            pnts += segment_points(pnts[-1], ang, turn, turn * r2, self.n)
            ang += turn
        # The last bend back to origin
        last_bend = ang + pi / 2
        if last_bend > 1e-13:
            r3 = pnts[-1].x / (cos(last_bend) - 1)
            pnts += segment_points(pnts[-1], ang, -last_bend, last_bend * r3, self.n)
            pnts += segment_points(pnts[-1], -pi / 2, 0.0, max(pnts[-1].y, 0.0), self.n)
        return pya.DPolygon(pnts)

    def finger_polygon(self, order_number):
        if self.finger_control <= order_number:  # The finger does not exist for given order_number.
            return None
        scale = self.finger_width + self.finger_gap
        x_max = max(self.finger_control, 1.0 / self.finger_control) * scale - self.finger_gap / 2
        trans = pya.DTrans(0, order_number % 2 == 1) * pya.DTrans(x_max, (order_number - 0.5) * scale)
        if self.finger_control <= 1.0:
            return self.t_poly(0.0, scale).transformed(trans)
        t_len = scale * pi / 2  # length of 90-degree turn segment
        s_len = scale * (2 * self.finger_control - 3)  # length of straight segment
        f_len = s_len + 2 * t_len  # total length of finger (including two 90-degree turns and straight)
        x = (self.finger_control - order_number) * f_len
        if x < t_len:  # The first turn is not full 90 degrees.
            return self.t_poly((x / t_len) * pi / 2, 0.0).transformed(trans)
        if s_len < 0.0:  # The first turn is limited by finger length. This only happens when order_number=0.
            return self.t_poly(pi / 2 - 2 * atan(-s_len / scale), -s_len).transformed(trans)
        if x < t_len + s_len:  # The straight segment is not full length.
            return self.t_poly(pi / 2, x - t_len).transformed(trans)
        if x < 2 * f_len - t_len:  # The straight segment is full, but the last turn does not exist yet.
            return self.t_poly(pi / 2, s_len).transformed(trans)
        if x < 2 * f_len:  # The last turn is below 90 degrees.
            return self.t_poly(pi * (1 + (x - 2 * f_len) / (2 * t_len)), s_len).transformed(trans)
        return self.t_poly(pi, s_len).transformed(trans)

    def super_smoothen_region(self, reg, r):
        rr = r / self.layout.dbu
        reg_mod = reg.sized(rr, 5).sized(-rr, 5).rounded_corners(rr, 0, self.n).rounded_corners(0, rr, self.n)
        reg += reg_mod
        return reg.smoothed(1)

    def get_finger_regions(self):
        # List of finger polygons
        i = 0
        polys = []
        while True:
            poly = self.finger_polygon(i)
            if poly is None:
                break
            polys.append(poly)
            i += 1

        # Create finger pad regions
        right_fingers = pya.Region([poly.to_itype(self.layout.dbu) for poly in polys])
        left_fingers = right_fingers.transformed(pya.Trans(2))

        return right_fingers, left_fingers

    def middle_gap_fill(self):
        scale = self.finger_width + self.finger_gap
        x_max = max(self.finger_control, 1.0 / self.finger_control) * scale - self.finger_gap / 2
        x_mid = x_max - self.finger_width
        y = scale / 2
        x = (x_mid + x_max) / 2
        l = 2 * x if self.finger_control < 1 else scale
        rr = (self.finger_width / 2 + self.ground_gap) / self.layout.dbu
        return (
            pya.Region(
                pya.DPolygon(
                    [pya.DPoint(-x, y), pya.DPoint(l - x, y), pya.DPoint(x, -y), pya.DPoint(x - l, -y)]
                ).to_itype(self.layout.dbu)
            )
            .sized(rr, 5)
            .rounded_corners(rr, rr, self.n)
        )

    def insert_wg_joint(self, reg, x0, xr, r):
        rr = r / self.layout.dbu
        reg += pya.Region(pya.DBox(xr, -r, 2 * x0 - xr, r).to_itype(self.layout.dbu)).rounded_corners(rr, rr, self.n)
        reg -= pya.Region(pya.DBox(x0, -r, 2 * x0 - xr, r).to_itype(self.layout.dbu))

    def build(self):
        # constants
        scale = self.finger_width + self.finger_gap
        x_max = max(self.finger_control, 1.0 / self.finger_control) * scale - self.finger_gap / 2
        x_mid = x_max - self.finger_width
        xport = x_max + self.ground_gap

        # Ground etch region
        right_fingers, left_fingers = self.get_finger_regions()
        region_ground = right_fingers + left_fingers
        region_ground.size(self.ground_gap / self.layout.dbu, 5)
        region_ground += self.middle_gap_fill()
        a2, b2 = eval_a2(self), eval_b2(self)
        self.insert_wg_joint(region_ground, xport, x_mid - self.ground_gap, b2 + a2 / 2)
        self.insert_wg_joint(region_ground, -xport, -x_mid + self.ground_gap, self.b + self.a / 2)
        region_ground = self.super_smoothen_region(region_ground, self.finger_gap + self.ground_gap)

        # Finalize finger pad regions
        self.insert_wg_joint(right_fingers, xport, x_mid, a2 / 2)
        self.insert_wg_joint(left_fingers, -xport, -x_mid, self.a / 2)
        right_fingers = self.super_smoothen_region(right_fingers, self.finger_gap)
        left_fingers = self.super_smoothen_region(left_fingers, self.finger_gap)

        # Insert waveguide segments in both ends, if fixed_length is set
        if self.fixed_length != 0:
            xfixed = self.fixed_length / 2
            if xfixed < xport:
                raise ValueError(f"SmoothCapacitor parameters not compatible with fixed_length={self.fixed_length}")
            region_ground += pya.Region(
                pya.DBox(xport, -b2 - a2 / 2, xfixed, b2 + a2 / 2).to_itype(self.layout.dbu)
            ) + pya.Region(
                pya.DBox(-xfixed, -self.b - self.a / 2, -xport, self.b + self.a / 2).to_itype(self.layout.dbu)
            )
        else:
            xfixed = xport
        # Always insert tolerance to secure trace connection
        right_fingers += pya.Region(pya.DBox(xport - 0.001, -a2 / 2, xfixed + 1, a2 / 2).to_itype(self.layout.dbu))
        left_fingers += pya.Region(
            pya.DBox(-xfixed - 1, -self.a / 2, -xport + 0.001, self.a / 2).to_itype(self.layout.dbu)
        )
        xport = xfixed

        # Create shapes into cell
        region = region_ground - right_fingers - left_fingers
        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(region)

        # protection
        region_protection = region_ground.sized(self.margin / self.layout.dbu, 5)
        self.add_protection(region_protection)

        # Add size into annotation layer
        self.cell.shapes(self.get_layer("annotations")).insert(pya.DText(str(round(self.finger_control, 5)), 0, 0))

        # Create ports
        self.add_port("a", pya.DPoint(-xport, 0), pya.DVector(-1, 0))
        self.add_port("b", pya.DPoint(xport, 0), pya.DVector(1, 0))

    @classmethod
    def get_sim_ports(cls, simulation):
        return FingerCapacitorSquare.get_sim_ports(simulation)
